'''
Created Date: Thursday June 6th 2024 +1000
Author: Peter Baker
-----
Last Modified: Thursday June 6th 2024 1:39:55 pm +1000
Modified By: Peter Baker
-----
Description: Registry API L3 module.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------

12-07-2024 | Parth Kulkarni | Adding some useful L3 methods to retrieve summarized info of registry items. 
28-06-2024 | Parth Kulkarni | Completion of Registry L3 interface with General, Admin and Other endpoints. 
18-06-2024 | Peter Baker | Initial proof of concept with fetch/update methods from L2. Sub Modules for each subtype.

'''

from provenaclient.auth.manager import AuthManager
from provenaclient.models.general import HealthCheckResponse
from provenaclient.utils.config import Config
from provenaclient.modules.module_helpers import *
from provenaclient.clients import RegistryClient
from ProvenaInterfaces.RegistryModels import *
from ProvenaInterfaces.RegistryAPI import *
from typing import Optional
from provenaclient.utils.helpers import convert_to_item_subtype, write_file_helper, get_and_validate_file_path



DEFAULT_CONFIG_FILE_NAME = "registry-api.env"

# L3 interface.

class RegistryAdminClient(ModuleService):
    _registry_client: RegistryClient

    def __init__(self, auth: AuthManager, config: Config, registry_client: RegistryClient) -> None:
        """
        Admin sub module of the Registry API providing functionality
        for the admin endpoints.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow
            method.
        config : Config
            A config object which contains information related to the Provena
            instance. 
        auth_client: AuthClient
            The instantiated auth client
        """
        self._auth = auth
        self._config = config

        # Clients related to the registry_api scoped as private.
        self._registry_client = registry_client

    
    async def export_items(self) -> RegistryExportResponse:
        """Provides a mechanism for admins to dump the current contents of the registry table without any validation/parsing.
        
        Returns
        -------
        RegistryExportResponse
            A status response including items in the payload
        """

        return await self._registry_client.admin.export_items()
    
    async def import_items(self, registry_import_request: RegistryImportRequest) -> RegistryImportResponse:
        """This admin only endpoint enables rapid restoration of items in into the registry table.

        Parameters
        ----------
        registry_import_request : RegistryImportRequest
            Contains the import mode, more info can be found on API docs.

        Returns
        -------
        RegistryImportResponse
            Returns an import response which includes status + statistics.
        """

        return await self._registry_client.admin.import_items(
            registry_import_request=registry_import_request
        )
    
    async def restore_items_from_dynamo_table(self, restore_request: RegistryRestoreRequest) -> RegistryImportResponse:
        """Provides an admin only mechanism for copying/restoring the contents from another dynamoDB table into the currently active registry table. 
        This endpoint does not create any new tables - it just uses the items from a restored table (e.g. from a backup) as inputs to an import operation 
        against the current registry table.

        Parameters
        ----------
        restore_request : RegistryRestoreRequest
            The restore request settings - these will be used when propagating the items from the external table.

        Returns
        -------
        RegistryImportResponse
            Returns information about the import, including status and statistics.
        """

        return await self._registry_client.admin.restore_items_from_dynamo_table(
            restore_request=restore_request
        )
    
    async def generate_config_file(self, required_only: bool = True, file_path: Optional[str] = None, write_to_file: bool = False) -> str:
        """Generates a nicely formatted .env file of the current required/non supplied properties 
        Used to quickly bootstrap a local environment or to understand currently deployed API.

        Parameters
        ----------
        required_only : bool, optional
            By default True
        file_path: str, optional
            The path you want to save the config file at WITH the file name. If you don't specify a path
            this will be saved in a relative directory.
        write_to_file: bool, By default False
            A boolean flag to indicate whether you want to save the config response to a file
            or not.

        Returns
        ----------
        str: Response containing the config text.

        """

        file_path = get_and_validate_file_path(file_path=file_path, write_to_file=write_to_file, default_file_name=DEFAULT_CONFIG_FILE_NAME)

        config_text: str = await self._registry_client.admin.generate_config_file(required_only=required_only)

        if config_text is None:
            raise ValueError(f"No data returned for generate config file endpoint.")

        # Write to file if config text is not None, write to file is True and file path is not None.
        if write_to_file:
            if file_path is None:
                raise ValueError("File path is not set for writing the CSV.")
            write_file_helper(file_path=file_path, content=config_text)

        return config_text
    
    async def delete(self, id: str, item_subtype: Optional[ItemSubType] = None) -> StatusResponse:
        """Admin only endpoint for deleting item from registry. USE CAREFULLY!

        Parameters
        ----------
        id : str
            ID of entity/item you want to delete.
        item_subtype : Optional[ItemSubType]
            Subtype of item you want to delete (E.g ORGANISATION, PERSON, CREATE)
            If not provided, it will be fetched from the registry.

        Returns
        -------
        StatusResponse
            Response indicating the success/failure of your request.
        """

        if item_subtype is None: 
            fetch_item = await self._registry_client.general.general_fetch_item(id=id)
            
            if fetch_item.item: 
                item_subtype_str: Optional[str] = fetch_item.item.get("item_subtype")
                item_subtype = convert_to_item_subtype(item_subtype_str)
            else:
                raise ValueError("Item not found")
                        
        return await self._registry_client.admin.delete_item(
            id = id, 
            item_subtype=item_subtype
        )
    
    
class RegistryBaseClass(ModuleService):
    _registry_client: RegistryClient

    def __init__(self, auth: AuthManager, config: Config, registry_client: RegistryClient, item_subtype: ItemSubType) -> None:
        """
        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance.
        registry_client : RegistryClient
            The registry client to use for registry interactions.
        item_subtype : ItemSubType
            The subtype of the item (e.g., ORGANISATION, PERSON).
        """
        self._auth = auth
        self._config = config
        self._registry_client = registry_client
        self.item_subtype = item_subtype

    
    async def admin_delete(self, id: str) -> StatusResponse:
        """Admin only endpoint for deleting item from registry. USE CAREFULLY!

        Parameters
        ----------
        id : str
            ID of entity/item you want to delete.

        Returns
        -------
        StatusResponse
            Response indicating the success/failure of your request.
        """

        return await self._registry_client.admin.delete_item(
            id = id, 
            item_subtype=self.item_subtype
        )

    async def revert_item(self, revert_request: ItemRevertRequest) -> ItemRevertResponse:
        """
        Reverts an item in the registry based on item subtype.

        Parameters
        ----------
        revert_request : ItemRevertRequest
            The revert request

        Returns
        -------
        ItemRevertResponse
            The revert response
        """
        return await self._registry_client.revert_item(
            revert_request=revert_request,
            item_subtype=self.item_subtype
        )
    
    async def get_schema(self) -> JsonSchemaResponse:
        """
        Gets the schema for the item subtype

        Returns
        -------
        JsonSchemaResponse
            The JSON schema response
        """
        return await self._registry_client.get_schema(
            item_subtype=self.item_subtype
        )
    
    async def evaluate_auth_access(self, id: str) -> DescribeAccessResponse:
        """
        Evaluates the auth access for an item based on item subtype.

        Parameters
        ----------
        id : str
            The item ID

        Returns
        -------
        DescribeAccessResponse
            The describe access response
        """
        return await self._registry_client.evaluate_auth_access(
            id=id, 
            item_subtype=self.item_subtype
        )
    
    async def get_auth_configuration(self, id: str) -> AccessSettings:
        """
        Gets the auth configuration for an item based on item subtype.

        Parameters
        ----------
        id : str
            The item ID

        Returns
        -------
        AccessSettings
            The access settings
        """
        return await self._registry_client.get_auth_configuration(
            id=id,
            item_subtype=self.item_subtype
        )
    
    async def modify_auth_configuration(self, id: str, auth_change_request: AccessSettings) -> StatusResponse:
        """
        Modifies the auth configuration for an item based on item subtype.

        Parameters
        ----------
        id : str
            The item ID
        auth_change_request : AccessSettings
            The auth change request

        Returns
        -------
        StatusResponse
            The status response
        """
        return await self._registry_client.modify_auth_configuration(
            id=id,
            auth_change_request=auth_change_request,
            item_subtype=self.item_subtype
        )
    
    async def get_auth_roles(self) -> AuthRolesResponse:
        """
        Gets the auth roles for the item subtype.

        Returns
        -------
        AuthRolesResponse
            The auth roles response
        """
        return await self._registry_client.get_auth_roles(
            item_subtype=self.item_subtype
        )
    
    async def lock_resource(self, lock_resource_request: LockChangeRequest) -> StatusResponse:
        """
        Locks a resource in the registry based on item subtype.

        Parameters
        ----------
        lock_resource_request : LockChangeRequest
            The lock resource request

        Returns
        -------
        StatusResponse
            The status response
        """
        return await self._registry_client.lock_resource(
            lock_resource_request=lock_resource_request,
            item_subtype=self.item_subtype
        )

    async def unlock_resource(self, unlock_resource_request: LockChangeRequest) -> StatusResponse:
        """
        Unlocks a resource in the registry based on item subtype.

        Parameters
        ----------
        unlock_resource_request : LockChangeRequest
            The unlock resource request

        Returns
        -------
        StatusResponse
            The status response
        """
        return await self._registry_client.unlock_resource(
            unlock_resource_request=unlock_resource_request,
            item_subtype=self.item_subtype
        )
    
    async def get_lock_history(self, id: str) -> LockHistoryResponse:
        """
        Gets the lock history for an item based on item subtype.

        Parameters
        ----------
        id : str
            The item ID

        Returns
        -------
        LockHistoryResponse
            The lock history response
        """
        return await self._registry_client.get_lock_history(
            handle_id=id,
            item_subtype=self.item_subtype
        )
    
    async def get_lock_status(self, id: str) -> LockStatusResponse:
        """
        Gets the lock status for an item based on item subtype.

        Parameters
        ----------
        id : str
            The item ID

        Returns
        -------
        LockStatusResponse
            The lock status response
        """
        return await self._registry_client.get_lock_status(
            id=id,
            item_subtype=self.item_subtype
        )

    
class OrganisationClient(RegistryBaseClass):
    _registry_client: RegistryClient

    def __init__(self, auth: AuthManager, config: Config, registry_client: RegistryClient) -> None:
        """
        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance.
        registry_client : RegistryClient
            The registry client to use for registry interactions.
        """
        super().__init__(auth=auth, config=config, registry_client=registry_client, item_subtype=ItemSubType.ORGANISATION)

    async def fetch(self, id: str, seed_allowed: Optional[bool] = None) -> OrganisationFetchResponse:
        """
        Fetches an organisation from the registry

        Args:
            id (str): The organisation ID
            seed_allowed (Optional[bool], optional): Allow seed items. Defaults to None.

        Returns:
            OrganisationFetchResponse: The fetch response
        """
        return await self._registry_client.fetch_item(
            id=id,
            item_subtype=self.item_subtype,
            fetch_response_model=OrganisationFetchResponse,
            seed_allowed=seed_allowed
        )

    async def update(self, id: str, domain_info: OrganisationDomainInfo, reason: Optional[str]) -> StatusResponse:
        """
        Updates an organisation in the registry

        Args:
            id (str): The id of the organisation
            domain_info (OrganisationDomainInfo): The new domain info
            reason (Optional[str]): The reason if any

        Returns:
            StatusResponse: Status response
        """
        return await self._registry_client.update_item(
            id=id,
            item_subtype=self.item_subtype,
            domain_info=domain_info,
            reason=reason,
            update_response_model=StatusResponse,
        )

    async def list_items(self, list_items_payload: GeneralListRequest) -> OrganisationListResponse:
        """
        Lists all organisations within the registry based on filter criteria.

        Parameters
        ----------
        list_items_payload : GeneralListRequest
            Payload containing the filter/sort criteria

        Returns
        -------
        OrganisationListResponse: The list response
        """
        return await self._registry_client.list_items(
            list_items_payload=list_items_payload,
            item_subtype=self.item_subtype, 
            update_model_response=OrganisationListResponse
        )

    async def seed_item(self) -> OrganisationSeedResponse:
        """
        Seeds an organisation in the registry

        Returns
        -------
        OrganisationSeedResponse: The seed response
        """
        return await self._registry_client.seed_item(
            item_subtype=self.item_subtype,
            seed_model_response=OrganisationSeedResponse
        )

    async def create_item(self, create_item_request: OrganisationDomainInfo) -> OrganisationCreateResponse:
        """
        Creates an organisation in the registry

        Parameters
        ----------
        create_item_request : OrganisationDomainInfo
            The create item request

        Returns
        -------
        OrganisationCreateResponse: The create response
        """
        return await self._registry_client.create_item(
            create_item_request=create_item_request,
            item_subtype=self.item_subtype,
            create_response_model=OrganisationCreateResponse
        )

    async def validate_item(self, validate_request: OrganisationDomainInfo) -> StatusResponse:
        """
        Validates an organisation in the registry

        Parameters
        ----------
        validate_request : OrganisationDomainInfo
            The validate request

        Returns
        -------
        StatusResponse: The status response
        """
        return await self._registry_client.validate_item(
            validate_request=validate_request,
            item_subtype=self.item_subtype
        )
    

class PersonClient(RegistryBaseClass):
    _registry_client: RegistryClient

    def __init__(self, auth: AuthManager, config: Config, registry_client: RegistryClient) -> None:
        """

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow
            method.
        config : Config
            A config object which contains information related to the Provena
            instance.
        """
        
        super().__init__(auth=auth, config=config, registry_client=registry_client, item_subtype=ItemSubType.PERSON)

    async def fetch(self, id: str, seed_allowed: Optional[bool] = None) -> PersonFetchResponse:
        """
        Fetches a person from the registry

        Args:
            id (str): The person ID
            seed_allowed (Optional[bool], optional): Allow seed items. Defaults to None.

        Returns:
            PersonFetch: The fetch response
        """
        return await self._registry_client.fetch_item(
            id=id,
            item_subtype=self.item_subtype,
            fetch_response_model=PersonFetchResponse,
            seed_allowed=seed_allowed
        )

    async def update(self, id: str, domain_info: PersonDomainInfo, reason: Optional[str]) -> StatusResponse:
        """
        Updates a person in the registry

        Args:
            id (str): The id of the organisation
            domain_info (OrganisationDomainInfo): The new domain info
            reason (Optional[str]): The reason if any

        Returns:
            StatusResponse: Status response
        """
        return await self._registry_client.update_item(
            id=id,
            item_subtype=self.item_subtype,
            domain_info=domain_info,
            reason=reason,
            update_response_model=StatusResponse,
        )
    
    async def list_items(self, list_items_payload: GeneralListRequest) -> PersonListResponse:
        """Lists all person(s) within registry based on filter
        criteria.

        Parameters
        ----------
        list_items_payload : GeneralListRequest
            Payload contaning the filter/sort criteria
        """

        return await self._registry_client.list_items(
            list_items_payload=list_items_payload,
            item_subtype=self.item_subtype, 
            update_model_response=PersonListResponse
        )
    
    async def seed_item(self) -> PersonSeedResponse:
        """
        Seeds a person in the registry

        Returns
        -------
        PersonSeedResponse: The seed response
        """

        return await self._registry_client.seed_item(
            item_subtype=self.item_subtype,
            seed_model_response=PersonSeedResponse
        )
    
    async def create_item(self, create_item_request: PersonDomainInfo) -> PersonCreateResponse:
        """
        Creates a person in the registry

        Parameters
        ----------
        create_item_request : OrganisationDomainInfo
            The create item request

        Returns
        -------
        PersonCreateResponse: The create response
        """

        return await self._registry_client.create_item(
            create_item_request=create_item_request,
            item_subtype=self.item_subtype,
            create_response_model=PersonCreateResponse
        )
    
    async def validate_item(self, validate_request: PersonDomainInfo) -> StatusResponse:
        """
        Validates a person in the registry

        Parameters
        ----------
        validate_request : PersonDomainInfo
            The validate request

        Returns
        -------
        StatusResponse: The status response
        """

        return await self._registry_client.validate_item(
            validate_request=validate_request,
            item_subtype=self.item_subtype
        )
    

class CreateActivityClient(RegistryBaseClass):
    _registry_client: RegistryClient

    def __init__(self, auth: AuthManager, config: Config, registry_client: RegistryClient) -> None:
        """
        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance.
        registry_client : RegistryClient
            The registry client to use for registry interactions.
        """
        
        super().__init__(auth=auth, config=config, registry_client=registry_client, item_subtype=ItemSubType.CREATE)

    async def fetch(self, id: str, seed_allowed: Optional[bool] = None) -> CreateFetchResponse:
        """
        Fetches a create activity item from the registry

        Args:
            id (str): The create activity item ID
            seed_allowed (Optional[bool], optional): Allow seed items. Defaults to None.

        Returns:
            CreateFetchResponse: The fetch response
        """
        return await self._registry_client.fetch_item(
            id=id,
            item_subtype=self.item_subtype,
            fetch_response_model=CreateFetchResponse,
            seed_allowed=seed_allowed
        )

    async def list_items(self, list_items_payload: GeneralListRequest) -> CreateListResponse:
        """
        Lists all create activity items within the registry based on filter criteria.

        Parameters
        ----------
        list_items_payload : GeneralListRequest
            Payload containing the filter/sort criteria

        Returns
        -------
        CreateListResponse: The list response
        """
        return await self._registry_client.list_items(
            list_items_payload=list_items_payload,
            item_subtype=self.item_subtype,
            update_model_response=CreateListResponse
        )
    
    async def validate_item(self, validate_request: CreateDomainInfo) -> StatusResponse:
        """
        Validates a create-item activity in the registry

        Parameters
        ----------
        validate_request : CreateDomainInfo
            The validate request

        Returns
        -------
        StatusResponse: The status response
        """

        return await self._registry_client.validate_item(
            validate_request=validate_request,
            item_subtype=self.item_subtype
        )



class VersionActivityClient(RegistryBaseClass):
    _registry_client: RegistryClient

    def __init__(self, auth: AuthManager, config: Config, registry_client: RegistryClient) -> None:
        """
        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance.
        registry_client : RegistryClient
            The registry client to use for registry interactions.
        """
        
        super().__init__(auth=auth, config=config, registry_client=registry_client, item_subtype=ItemSubType.VERSION)

    async def fetch(self, id: str, seed_allowed: Optional[bool] = None) -> VersionFetchResponse:
        """
        Fetches a version activity item from the registry

        Args:
            id (str): The version activity item ID
            seed_allowed (Optional[bool], optional): Allow seed items. Defaults to None.

        Returns:
            VersionFetchResponse: The fetch response
        """
        return await self._registry_client.fetch_item(
            id=id,
            item_subtype=self.item_subtype,
            fetch_response_model=VersionFetchResponse,
            seed_allowed=seed_allowed
        )

    async def list_items(self, list_items_payload: GeneralListRequest) -> VersionListResponse:
        """
        Lists all version activity items within the registry based on filter criteria.

        Parameters
        ----------
        list_items_payload : GeneralListRequest
            Payload containing the filter/sort criteria

        Returns
        -------
        VersionListResponse: The list response
        """
        return await self._registry_client.list_items(
            list_items_payload=list_items_payload,
            item_subtype=self.item_subtype,
            update_model_response=VersionListResponse
        )
    
    async def validate_item(self, validate_request: VersionDomainInfo) -> StatusResponse:
        """
        Validates a version activity in the registry

        Parameters
        ----------
        validate_request : VersionDomainInfo
            The validate request

        Returns
        -------
        StatusResponse: The status response
        """

        return await self._registry_client.validate_item(
            validate_request=validate_request,
            item_subtype=self.item_subtype
        )
    

class ModelRunActivityClient(RegistryBaseClass):
    _registry_client: RegistryClient

    def __init__(self, auth: AuthManager, config: Config, registry_client: RegistryClient) -> None:
        """
        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance.
        registry_client : RegistryClient
            The registry client to use for registry interactions.
        """
        
        super().__init__(auth=auth, config=config, registry_client=registry_client, item_subtype=ItemSubType.MODEL_RUN)

    async def fetch(self, id: str, seed_allowed: Optional[bool] = None) -> ModelRunFetchResponse:
        """
        Fetches a model run activity item from the registry

        Args:
            id (str): The model run activity item ID
            seed_allowed (Optional[bool], optional): Allow seed items. Defaults to None.

        Returns:
            ModelRunFetchResponse: The fetch response
        """
        return await self._registry_client.fetch_item(
            id=id,
            item_subtype=self.item_subtype,
            fetch_response_model=ModelRunFetchResponse,
            seed_allowed=seed_allowed
        )

    async def list_items(self, list_items_payload: GeneralListRequest) -> ModelRunListResponse:
        """
        Lists all model run activity items within the registry based on filter criteria.

        Parameters
        ----------
        list_items_payload : GeneralListRequest
            Payload containing the filter/sort criteria

        Returns
        -------
        ModelRunListResponse: The list response
        """
        return await self._registry_client.list_items(
            list_items_payload=list_items_payload,
            item_subtype=self.item_subtype,
            update_model_response=ModelRunListResponse
        )
    
    async def validate_item(self, validate_request: ModelRunDomainInfo) -> StatusResponse:
        """
        Validates a model run in the registry

        Parameters
        ----------
        validate_request : ModelRunDomainInfo
            The validate request

        Returns
        -------
        StatusResponse: The status response
        """

        return await self._registry_client.validate_item(
            validate_request=validate_request,
            item_subtype=self.item_subtype
        )
    

    
class ModelClient(RegistryBaseClass):
    _registry_client: RegistryClient

    def __init__(self, auth: AuthManager, config: Config, registry_client: RegistryClient) -> None:
        """

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow
            method.
        config : Config
            A config object which contains information related to the Provena
            instance.
        """

        super().__init__(auth=auth, config=config, registry_client=registry_client, item_subtype=ItemSubType.MODEL)

    async def fetch(self, id: str, seed_allowed: Optional[bool] = None) -> ModelFetchResponse:
        """
        Fetches a model from the registry

        Args:
            id (str): The model ID
            seed_allowed (Optional[bool], optional): Allow seed items. Defaults to None.

        Returns:
            ModelFetchResponse: The fetch response
        """
        return await self._registry_client.fetch_item(
            id=id,
            item_subtype=self.item_subtype,
            fetch_response_model=ModelFetchResponse,
            seed_allowed=seed_allowed
        )
        
    async def update(self, id: str, domain_info: ModelDomainInfo, reason: Optional[str]) -> StatusResponse:
        """
        Updates a model in the registry

        Args:
            id (str): The id of the model
            domain_info (ModelDomainInfo): The new domain info
            reason (Optional[str]): The reason if any

        Returns:
            StatusResponse: Status response
        """
        return await self._registry_client.update_item(
            id=id,
            item_subtype=self.item_subtype,
            domain_info=domain_info,
            reason=reason,
            update_response_model=StatusResponse,
        )
    
    async def list_items(self, list_items_payload: GeneralListRequest) -> ModelListResponse:
        """Lists all model(s) within registry based on filter
        criteria.

        Parameters
        ----------
        list_items_payload : GeneralListRequest
            Payload contaning the filter/sort criteria
        """

        return await self._registry_client.list_items(
            list_items_payload=list_items_payload,
            item_subtype=self.item_subtype, 
            update_model_response=ModelListResponse
        )
    
    async def seed_item(self) -> ModelSeedResponse:
        """
        Seeds a model in the registry

        Returns
        -------
        ModelSeedResponse: The seed response
        """

        return await self._registry_client.seed_item(
            item_subtype=self.item_subtype,
            seed_model_response=ModelSeedResponse
        )
    
    
    async def create_item(self, create_item_request: ModelDomainInfo) -> ModelCreateResponse:
        """
        Creates a model in the registry

        Parameters
        ----------
        create_item_request : ModelDomainInfo
            The create item request

        Returns
        -------
        ModelCreateResponse: The create response
        """

        return await self._registry_client.create_item(
            create_item_request=create_item_request,
            item_subtype=self.item_subtype,
            create_response_model=ModelCreateResponse
        )
    
    async def validate_item(self, validate_request: ModelDomainInfo) -> StatusResponse:
        """
        Validates a model item in the registry

        Parameters
        ----------
        validate_request : ModelDomainInfo
            The validate request

        Returns
        -------
        StatusResponse: The status response
        """

        return await self._registry_client.validate_item(
            validate_request=validate_request,
            item_subtype=self.item_subtype
        )
    
     
    async def version_item(self, version_request: VersionRequest) -> VersionResponse:
        """
        Versions a model in the registry

        Parameters
        ----------
        version_request : VersionRequest
            The version request

        Returns
        -------
        VersionResponse: The version response
        """

        return await self._registry_client.version(
            version_request=version_request,
            item_subtype=self.item_subtype
        )
    
class ModelRunWorkFlowClient(RegistryBaseClass):
    _registry_client: RegistryClient

    def __init__(self, auth: AuthManager, config: Config, registry_client: RegistryClient) -> None:
        """

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow
            method.
        config : Config
            A config object which contains information related to the Provena
            instance.
        """
        
        super().__init__(auth=auth, config=config, registry_client=registry_client, item_subtype=ItemSubType.WORKFLOW_TEMPLATE)

    async def fetch(self, id: str, seed_allowed: Optional[bool] = None) -> ModelRunWorkflowTemplateFetchResponse:
        """
        Fetches a model run workflow template from the registry

        Args:
            id (str): The model run workflow template ID
            seed_allowed (Optional[bool], optional): Allow seed items. Defaults to None.

        Returns:
            ModelFetchResponse: The fetch response
        """
        return await self._registry_client.fetch_item(
            id=id,
            item_subtype=self.item_subtype,
            fetch_response_model=ModelRunWorkflowTemplateFetchResponse,
            seed_allowed=seed_allowed
        )
        
    async def update(self, id: str, domain_info: ModelRunWorkflowTemplateDomainInfo, reason: Optional[str]) -> StatusResponse:
        """
        Updates a model in the registry

        Args:
            id (str): The id of the model
            domain_info (ModelDomainInfo): The new domain info
            reason (Optional[str]): The reason if any

        Returns:
            StatusResponse: Status response
        """
        return await self._registry_client.update_item(
            id=id,
            item_subtype=self.item_subtype,
            domain_info=domain_info,
            reason=reason,
            update_response_model=StatusResponse,
        )
    
    async def list_items(self, list_items_payload: GeneralListRequest) -> ModelRunWorkflowTemplateListResponse:
        """Lists all model(s) within registry based on filter
        criteria.

        Parameters
        ----------
        list_items_payload : GeneralListRequest
            Payload contaning the filter/sort criteria
        """

        return await self._registry_client.list_items(
            list_items_payload=list_items_payload,
            item_subtype=self.item_subtype,
            update_model_response=ModelRunWorkflowTemplateListResponse
        )
    
    async def seed_item(self) -> ModelRunWorkflowTemplateSeedResponse:
        """
        Seeds a model run workflow template in the registry

        Returns
        -------
        ModelRunWorkflowTemplateSeedResponse: The seed response
        """

        return await self._registry_client.seed_item(
            item_subtype=self.item_subtype,
            seed_model_response=ModelRunWorkflowTemplateSeedResponse
        )
    
    
    async def create_item(self, create_item_request: ModelRunWorkflowTemplateDomainInfo) -> ModelRunWorkflowTemplateCreateResponse:
        """
        Creates a model run workflow template in the registry

        Parameters
        ----------
        create_item_request : ModelRunWorkflowTemplateDomainInfo
            The create item request

        Returns
        -------
        ModelRunWorkflowTemplateCreateResponse: The create response
        """

        return await self._registry_client.create_item(
            create_item_request=create_item_request,
            item_subtype=self.item_subtype,
            create_response_model=ModelRunWorkflowTemplateCreateResponse
        )
    
    async def validate_item(self, validate_request: ModelRunWorkflowTemplateDomainInfo) -> StatusResponse:
        """
        Validates a model run workflow template item in the registry

        Parameters
        ----------
        validate_request : ModelRunWorkflowTemplateDomainInfo
            The validate request

        Returns
        -------
        StatusResponse: The status response
        """

        return await self._registry_client.validate_item(
            validate_request=validate_request,
            item_subtype=self.item_subtype
        )
    
    async def version_item(self, version_request: VersionRequest) -> VersionResponse:
        """
        Versions a model run workflow template in the registry

        Parameters
        ----------
        version_request : VersionRequest
            The version request

        Returns
        -------
        VersionResponse: The version response
        """

        return await self._registry_client.version(
            version_request=version_request,
            item_subtype=self.item_subtype
        )
    

class DatasetTemplateClient(RegistryBaseClass):
    _registry_client: RegistryClient

    def __init__(self, auth: AuthManager, config: Config, registry_client: RegistryClient) -> None:
        """
        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow
            method.
        config : Config
            A config object which contains information related to the Provena
            instance.
        """
        
        super().__init__(auth=auth, config=config, registry_client=registry_client, item_subtype=ItemSubType.DATASET_TEMPLATE)

    async def fetch(self, id: str, seed_allowed: Optional[bool] = None) -> DatasetTemplateFetchResponse:
        """
        Fetches a dataset template from the registry

        Args:
            id (str): The dataset template ID
            seed_allowed (Optional[bool], optional): Allow seed items. Defaults to None.

        Returns:
            DatasetTemplateFetchResponse: The fetch response
        """
        return await self._registry_client.fetch_item(
            id=id,
            item_subtype=self.item_subtype,
            fetch_response_model=DatasetTemplateFetchResponse,
            seed_allowed=seed_allowed
        )

    async def update(self, id: str, domain_info: DatasetTemplateDomainInfo, reason: Optional[str]) -> StatusResponse:
        """
        Updates a dataset template in the registry

        Args:
            id (str): The id of the dataset template
            domain_info (DatasetTemplateDomainInfo): The new domain info
            reason (Optional[str]): The reason if any

        Returns:
            StatusResponse: Status response
        """
        return await self._registry_client.update_item(
            id=id,
            item_subtype=self.item_subtype,
            domain_info=domain_info,
            reason=reason,
            update_response_model=StatusResponse,
        )

    async def list_items(self, list_items_payload: GeneralListRequest) -> DatasetTemplateListResponse:
        """
        Lists all dataset templates within the registry based on filter
        criteria.

        Parameters
        ----------
        list_items_payload : GeneralListRequest
            Payload containing the filter/sort criteria
        """
        return await self._registry_client.list_items(
            list_items_payload=list_items_payload,
            item_subtype=self.item_subtype,
            update_model_response=DatasetTemplateListResponse
        )

    async def seed_item(self) -> DatasetTemplateSeedResponse:
        """
        Seeds a dataset template in the registry

        Returns
        -------
        DatasetTemplateSeedResponse: The seed response
        """
        return await self._registry_client.seed_item(
            item_subtype=self.item_subtype,
            seed_model_response=DatasetTemplateSeedResponse
        )

    async def create_item(self, create_item_request: DatasetTemplateDomainInfo) -> DatasetTemplateCreateResponse:
        """
        Creates a dataset template in the registry

        Parameters
        ----------
        create_item_request : DatasetTemplateDomainInfo
            The create item request

        Returns
        -------
        DatasetTemplateCreateResponse: The create response
        """
        return await self._registry_client.create_item(
            create_item_request=create_item_request,
            item_subtype=self.item_subtype,
            create_response_model=DatasetTemplateCreateResponse
        )
    
    async def validate_item(self, validate_request: DatasetTemplateDomainInfo) -> StatusResponse:
        """
        Validates a dataset template item in the registry

        Parameters
        ----------
        validate_request : DatasetTemplateDomainInfo
            The validate request

        Returns
        -------
        StatusResponse: The status response
        """

        return await self._registry_client.validate_item(
            validate_request=validate_request,
            item_subtype=self.item_subtype
        )

    async def version_item(self, version_request: VersionRequest) -> VersionResponse:
        """
        Versions a dataset template in the registry

        Parameters
        ----------
        version_request : VersionRequest
            The version request

        Returns
        -------
        VersionResponse: The version response
        """
        return await self._registry_client.version(
            version_request=version_request,
            item_subtype=self.item_subtype
        )


class DatasetClient(RegistryBaseClass):
    _registry_client: RegistryClient

    def __init__(self, auth: AuthManager, config: Config, registry_client: RegistryClient) -> None:
        """
        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance.
        registry_client : RegistryClient
            The registry client to use for registry interactions.
        """
        
        super().__init__(auth=auth, config=config, registry_client=registry_client, item_subtype=ItemSubType.DATASET)

    async def fetch(self, id: str, seed_allowed: Optional[bool] = None) -> DatasetFetchResponse:
        """
        Fetches a dataset from the registry

        Args:
            id (str): The dataset ID
            seed_allowed (Optional[bool], optional): Allow seed items. Defaults to None.

        Returns:
            DatasetFetchResponse: The fetch response
        """
        return await self._registry_client.fetch_item(
            id=id,
            item_subtype=self.item_subtype,
            fetch_response_model=DatasetFetchResponse,
            seed_allowed=seed_allowed
        )

    async def list_items(self, list_items_payload: GeneralListRequest) -> DatasetListResponse:
        """
        Lists all datasets within the registry based on filter criteria.

        Parameters
        ----------
        list_items_payload : GeneralListRequest
            Payload containing the filter/sort criteria

        Returns
        -------
        DatasetListResponse: The list response
        """
        return await self._registry_client.list_items(
            list_items_payload=list_items_payload,
            item_subtype=self.item_subtype,
            update_model_response=DatasetListResponse
        )
    

    async def validate_item(self, validate_request: DatasetDomainInfo) -> StatusResponse:
        """
        Validates a dataset item in the registry

        Parameters
        ----------
        validate_request : DatasetDomainInfo
            The validate request

        Returns
        -------
        StatusResponse: The status response
        """

        return await self._registry_client.validate_item(
            validate_request=validate_request,
            item_subtype=self.item_subtype
        )


class StudyClient(RegistryBaseClass):
    _registry_client: RegistryClient

    def __init__(self, auth: AuthManager, config: Config, registry_client: RegistryClient) -> None:
        """
        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance.
        registry_client : RegistryClient
            The registry client to use for registry interactions.
        """
        
        super().__init__(auth=auth, config=config, registry_client=registry_client, item_subtype=ItemSubType.STUDY)

    async def fetch(self, id: str, seed_allowed: Optional[bool] = None) -> StudyFetchResponse:
        """
        Fetches a study from the registry

        Args:
            id (str): The study ID
            seed_allowed (Optional[bool], optional): Allow seed items. Defaults to None.

        Returns:
            StudyFetchResponse: The fetch response
        """
        return await self._registry_client.fetch_item(
            id=id,
            item_subtype=self.item_subtype,
            fetch_response_model=StudyFetchResponse,
            seed_allowed=seed_allowed
        )

    async def list_items(self, list_items_payload: GeneralListRequest) -> StudyListResponse:
        """
        Lists all studies within the registry based on filter criteria.

        Parameters
        ----------
        list_items_payload : GeneralListRequest
            Payload containing the filter/sort criteria

        Returns
        -------
        StudyListResponse: The list response
        """
        return await self._registry_client.list_items(
            list_items_payload=list_items_payload,
            item_subtype=self.item_subtype,
            update_model_response=StudyListResponse
        )

    async def seed_item(self) -> StudySeedResponse:
        """
        Seeds a study in the registry

        Returns
        -------
        StudySeedResponse: The seed response
        """
        return await self._registry_client.seed_item(
            item_subtype=self.item_subtype,
            seed_model_response=StudySeedResponse
        )

    async def update(self, id: str, domain_info: StudyDomainInfo, reason: Optional[str]) -> StatusResponse:
        """
        Updates a study in the registry

        Parameters
        ----------
        id : str
            The id of the study
        domain_info : StudyDomainInfo
            The new domain info
        reason : Optional[str]
            The reason, if any

        Returns
        -------
        StatusResponse: Status response
        """
        return await self._registry_client.update_item(
            id=id,
            item_subtype=self.item_subtype,
            domain_info=domain_info,
            reason=reason,
            update_response_model=StatusResponse,
        )

    async def create_item(self, create_item_request: StudyDomainInfo) -> StudyCreateResponse:
        """
        Creates a study in the registry

        Parameters
        ----------
        create_item_request : StudyDomainInfo
            The create item request

        Returns
        -------
        StudyCreateResponse: The create response
        """
        return await self._registry_client.create_item(
            create_item_request=create_item_request,
            item_subtype=self.item_subtype,
            create_response_model=StudyCreateResponse
        )
    
    async def validate_item(self, validate_request: StudyDomainInfo) -> StatusResponse:
        """
        Validates a study item in the registry

        Parameters
        ----------
        validate_request : StudyDomainInfo
            The validate request

        Returns
        -------
        StatusResponse: The status response
        """

        return await self._registry_client.validate_item(
            validate_request=validate_request,
            item_subtype=self.item_subtype
        )

    
class Registry(ModuleService):
    # L2 clients used
    _registry_client: RegistryClient

    # Admin sub module
    admin: RegistryAdminClient

    # Sub modules
    organisation: OrganisationClient
    person: PersonClient
    model: ModelClient
    model_run_workflow: ModelRunWorkFlowClient
    dataset_template: DatasetTemplateClient
    dataset: DatasetClient
    study: StudyClient
    create_activity: CreateActivityClient
    version_acitvity: VersionActivityClient
    model_run: ModelRunActivityClient
    
    def __init__(self, auth: AuthManager, config: Config, registry_client: RegistryClient) -> None:
        """

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow
            method.
        config : Config
            A config object which contains information related to the Provena
            instance.
        """
        # Module service
        self._auth = auth
        self._config = config

        # Clients related to the registry scoped as private.
        self._registry_client = registry_client

        # Admin sub module
        self.admin = RegistryAdminClient(
            auth=auth, config=config, registry_client=registry_client
        )

        # Sub modules
        self.organisation = OrganisationClient(
            auth=auth, config=config, registry_client=registry_client)    
        self.person = PersonClient(
            auth=auth, config=config, registry_client=registry_client
        )
        self.model = ModelClient(
            auth=auth, config=config, registry_client=registry_client)
        self.model_run_workflow = ModelRunWorkFlowClient(
            auth=auth, config=config, registry_client=registry_client
        )
        self.dataset_template = DatasetTemplateClient(
            auth=auth, config=config, registry_client=registry_client
        )
        self.study = StudyClient(
            auth=auth, config=config, registry_client=registry_client
        )
        self.create_activity = CreateActivityClient(
            auth=auth, config=config, registry_client=registry_client)
        self.version_acitvity = VersionActivityClient(
            auth=auth, config=config, registry_client=registry_client)
        self.model_run = ModelRunActivityClient(
            auth=auth, config=config, registry_client=registry_client
        )
    
    async def get_health_check(self) -> HealthCheckResponse:
        """
        Health check the API

        Returns:
            HealthCheckResponse: Response
        """
        return await self._registry_client.get_health_check()
    

    async def list_general_registry_items(self, general_list_request: GeneralListRequest) -> PaginatedListResponse:
        """
        Lists general registry items based on filter criteria.

        Parameters
        ----------
        general_list_request : GeneralListRequest
            The request containing filter and sort criteria.

        Returns
        -------
        PaginatedListResponse
            The response containing the paginated list of registry items.
        """

        return await self._registry_client.general.list_general_registry_items(
            general_list_request=general_list_request
        )
    
    async def list_registry_items_with_count(self) -> Dict[str, int]:
        """
        Retrieves a count of items in the registry, grouped by their subtypes.

        This method sends requests to list items in the registry and counts the number of items for each subtype.
        It handles pagination to ensure all items are counted.

        Returns
        -------
        Dict[str, int]
            A dictionary where the keys are item subtypes(string) and the values are the count of items for each subtype.
        """
    
        item_count: Dict[str, int] = {}

        general_list_request = GeneralListRequest(
            filter_by=FilterOptions(
                record_type=QueryRecordTypes.COMPLETE_ONLY,
                item_subtype=None, 
                release_reviewer=None,
                release_status=None
            ),
            sort_by=SortOptions(sort_type=None, ascending=False, begins_with=None),
            pagination_key=None, 
            page_size=20
        )

        while True:
            list_response = await self.list_general_registry_items(general_list_request=general_list_request)

            assert list_response.items is not None, f"Expected a list of items but none was present."

            for item in list_response.items:
                subtype: str = item.get("item_subtype")

                if not subtype:
                    unknown = "UNKNOWN"
    
                    if unknown not in item_count:
                        item_count[unknown] = 1
                    
                    else:
                        item_count[unknown] += 1
                    
                    continue  # Skip this current iteration if subtype is missing

                if subtype not in item_count:
                    item_count[subtype] = 1
                else:
                    item_count[subtype] += 1

            if not list_response.pagination_key:
                break

            general_list_request.pagination_key = list_response.pagination_key

        return item_count
    
    async def general_fetch_item(self, id: str) -> UntypedFetchResponse:
        """
        Fetches a general item from the registry.

        Parameters
        ----------
        id : str
            The ID of the item to fetch.

        Returns
        -------
        UntypedFetchResponse
            The fetch response containing the item details.
        """

        return await self._registry_client.general.general_fetch_item(
            id=id
        )

    
    async def get_current_provena_version(self) -> VersionResponse:
        """
        Gets the current Provena version deployed on your domain.

        Returns
        -------
        VersionResponse
            The response containing the current Provena version.
        """

        return await self._registry_client.general.get_current_provena_version()



'''
Created Date: Thursday May 30th 2024 +1000
Author: Peter Baker
-----
Last Modified: Thursday May 30th 2024 10:17:36 am +1000
Modified By: Peter Baker
-----
Description: A template file for creating an L3 module.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

# from provenaclient.auth.manager import AuthManager
# from provenaclient.utils.config import Config
# from pydantic import BaseModel
#
# L3 interface.
#
# class REPLACE(ModuleService):
#    _replace_client: REPLACEClient
#
#    def __init__(self, auth: AuthManager, config: Config, replace_client: REPLACEClient) -> None:
#        """
#
#        TODO
#
#        Parameters
#        ----------
#        auth : AuthManager
#            An abstract interface containing the user's requested auth flow
#            method.
#        config : Config
#            A config object which contains information related to the Provena
#            instance.
#        """
#        # Module service
#        self._auth = auth
#        self._config = config
#
#        # Clients related to the datastore scoped as private.
#        self._replace_client = replace_client
#
#    async def example_method(self) -> BaseModel:
#        return await self._replace_client.example()
#



'''
Created Date: Tuesday June 18th 2024 +1000
Author: Peter Baker
-----
Last Modified: Tuesday June 18th 2024 12:44:03 pm +1000
Modified By: Peter Baker
-----
Description: Implementations of the Auth interface defined in auth/manager.py
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

from typing import Any, Dict, Optional
from provenaclient.auth.manager import AuthManager, LogType
import requests
import webbrowser
import time
import os
from provenaclient.auth.helpers import AccessToken, Tokens, keycloak_refresh_token_request, validate_access_token, retrieve_keycloak_public_key
import json
from provenaclient.auth.manager import DEFAULT_LOG_LEVEL
from provenaclient.utils.config import Config


class DeviceFlow(AuthManager):
    keycloak_endpoint: str
    client_id: str
    scopes: list
    device_endpoint: str
    token_endpoint: str

    def __init__(self, config: Config, client_id: str, log_level: Optional[LogType] = None) -> None:
        f""" Create and generate a DeviceFlow object. The tokens are automatically refreshed when
        accessed through the get_auth() function.

        Tokens are cached in local storage with a configurable file name and are
        only reproduced if the refresh token expires.

        Parameters
        ----------
        keycloak_endpoint : str
            The keycloak endpoint to use.
        client_id : str
            The client id for the keycloak authorisation.
        log_level: Optional[LogType]
            The logging level to use - defaults to {DEFAULT_LOG_LEVEL} 
        """

        # construct parent class and include log level
        super().__init__(log_level=log_level)

        self.keycloak_endpoint = config.keycloak_endpoint
        self.client_id = client_id
        self.scopes: list = []
        self.file_name = ".tokens.json"
        self.device_endpoint = f'{self.keycloak_endpoint}/protocol/openid-connect/auth/device'
        self.token_endpoint = f'{self.keycloak_endpoint}/protocol/openid-connect/token'

        try:
            # First thing to do here is obtain the keycloak public key.
            self.public_key = retrieve_keycloak_public_key(
                logger=self.logger,
                keycloak_endpoint=self.keycloak_endpoint,
            )

        except Exception as e:
            raise Exception(
                "Failed to retrieve the Keycloak public key, authentication cannot proceed.") from e

        # Second thing will be to check if the tokens.json file is already present or not.
        # If it's present validate it, if fails then refresh else not present then fetch new tokens.
        if os.path.exists(self.file_name):
            self.tokens = self.load_tokens()
            if not self.tokens:
                self.logger.info(
                    "No tokens found or failed to load tokens.")
                self.start_device_flow()
            else:
                # Attempt to validate tokens
                if validate_access_token(
                    logger=self.logger,
                    access_token=self.tokens.access_token,
                    public_key=self.public_key
                ):
                    self.logger.info("Tokens are valid...")
                else:
                    try:
                        self.refresh_tokens()
                    except Exception as e:
                        self.logger.info(
                            f"Refresh token has expired or is invalid {e}")
                        self.start_device_flow()
        else:
            self.logger.info("No token file found, starting device flow.")
            self.start_device_flow()

    def get_token(self) -> str:
        """
        IMPLEMENTS BASE METHOD 

        Uses the current token - validates it,
        refreshes if necessary, and returns the valid token
        ready to be used.

        Returns
        -------
        str
            The access token

        Raises
        ------
        Exception
            Raises exception if tokens/public_key are not setup - make sure
            that the object is instantiated properly before calling this function.
        Exception
            If the token is invalid and cannot be refreshed.
        Exception
            If the token validation still fails after re-conducting the device flow.
        """

        if self.tokens is None or self.public_key is None:
            raise Exception(
                "Cannot generate token without access token or public key.")

        try:
            # Attempt to validate the current token.
            if validate_access_token(public_key=self.public_key, access_token=self.tokens.access_token, logger=self.logger):
                return self.tokens.access_token
            
            # didnt return, refresh and try again. 
            self.logger.info("Token was invalid. Attempting Refresh")
            self.refresh_tokens()
            if validate_access_token(public_key=self.public_key, access_token=self.tokens.access_token, logger=self.logger):
                return self.tokens.access_token
            
            # still no good, restart flow
            self.logger.info("Token was invalid after refresh. Re-iniating Device Flow")
            self.start_device_flow()
            if validate_access_token(public_key=self.public_key, access_token=self.tokens.access_token, logger=self.logger):
                return self.tokens.access_token

        except Exception as e:
            self.logger.info("Something went wrong during get_token operation. Starting device flow.")
            self.start_device_flow()
            if validate_access_token(public_key=self.public_key, access_token=self.tokens.access_token, logger=self.logger):
                return self.tokens.access_token
        
        # no error, but also no valid token. Something is wrong.
        err_msg = "Failed to obtain a valid token after refreshing and initiating a new device flow."
        self.logger.error(err_msg)
        raise Exception(err_msg)
            

    def force_refresh(self) -> None:
        """
        IMPLEMENTS BASE METHOD 
        A method to reset the current authentication state.
        """

        # Force refresh everything hear, so reset the tokens file and re-generate the device flow.
        self.clear_token_storage()
        self.start_device_flow()

    def refresh_tokens(self) -> None:
        """Attempts to refresh the authentication tokens using a stored refresh token. This method
        updates the current tokens if the refresh is successful.

        Raises
        ------
        ValueError
            If no initial tokens are set, indicating that there is nothing to refresh. 
        ValueError
            If the refresh operation fails due to missing access or refresh tokens in the response,
            suggesting a failure in the refresh process.
        """

        if self.tokens is None:
            raise ValueError(
                "Token refresh attempted with no initial tokens set. ")

        self.logger.info("Refreshing using refresh token")

        refreshed: Dict[str, Any] = self.make_token_refresh_request()

        access_token = refreshed.get('access_token')
        refresh_token = refreshed.get('refresh_token')

        if not access_token or not refresh_token:
            error_message = "Failed to refresh tokens: Missing access or refresh tokens"
            self.logger.info(error_message)
            raise ValueError(error_message)
        else:
            self.tokens = Tokens(
                access_token=access_token,
                refresh_token=refresh_token
            )
            self.save_tokens(self.tokens)

    def save_tokens(self, tokens: Tokens) -> None:
        """Saves authentication tokens to a local file in JSON format.

        Parameters
        ----------
        tokens : Tokens
            An object representing the authentication tokens containing 
            the access and refresh tokens.

        Raises
        -------
        Generic Exception
            A generic exception is raised that handles errors from IO/File operations.

        """
        self.logger.info("Saving tokens into local storage.")

        try:

            with open(self.file_name, 'w') as file:
                json.dump(tokens.dict(), file)
                self.logger.info("Tokens saved to file successfully.")

        except Exception as e:
            print(f"Failed to save tokens: {e}")

    def clear_token_storage(self) -> None:
        """Checks if the tokens.json file exists and accordingly removes it and resets
        token object saved to class variable.
        """
        if os.path.exists("tokens.json"):
            os.remove("tokens.json")
            self.logger.info("Stored tokens have been clear.")

        self.tokens = None

    def load_tokens(self) -> Optional[Tokens]:
        """Loads authentication tokens from a local JSON file and returns them as a Tokens object.

        Returns
        -------
        Tokens
            An object representing the authentication tokens containing 
            the access and refresh tokens.

        Raises
        -------
        Generic Exception
            A generic exception is raised that handles errors from IO/File operations.

        """

        self.logger.info("Looking for existing tokens in local storage.")

        try:
            with open(self.file_name, 'r') as file:
                token_data = json.load(file)
                return Tokens(**token_data)
        except Exception as e:
            print(f"Failed to load tokens: {e}")
            return None

    def make_token_refresh_request(self, tokens: Optional[Tokens] = None) -> Dict[str, Any]:
        """Performs the token refresh by making an HTTP post request to the token endpoint
        to obtain new access and refresh tokens.

        Parameters
        ----------
        tokens : Optional[Tokens], optional
            An optional Tokens object containing the refresh token. If not provided,
            the method will use the class variable stored tokens. 

            By default this parameter is None.

        Returns
        -------
        Dict[str, Any]
            A dictionary containing the new access and refresh tokens if the refresh is successful.

        Raises
        ------
        ValueError
            If no refresh token is provided or found in the class token variable. 
        Exception
            If the HTTP request fails a message is displayed with the HTTP status code. Can occur 
            if the refresh token has expired.
        """

        # make sure we have tokens to use
        desired_tokens: Optional[Tokens]
        if tokens:
            desired_tokens = tokens
        else:
            desired_tokens = self.tokens

        if not desired_tokens or not desired_tokens.refresh_token:
            raise ValueError("Refresh token is required but was not provided.")

        return keycloak_refresh_token_request(
            logger=self.logger,
            client_id=self.client_id,
            refresh_token=desired_tokens.refresh_token,
            scopes=self.scopes,
            token_endpoint=self.token_endpoint
        )

    def start_device_flow(self) -> None:
        """Initiates the device authorisation flow by requesting a device code from server and prompts
        user for authentication through the web browser and continues to handle the flow.

        Raises
        ------
        Exception
            If the request to the server fails or if the server response is not of status code 200,
            suggesting that the flow could not initiated.
        """

        self.logger.info("Initiating auth device flow.")

        data = {
            "client_id": self.client_id,
            "scopes": ' '.join(self.scopes)
        }

        response = requests.post(self.device_endpoint, data=data)

        if response.status_code == 200:
            response_data = response.json()
            self.device_code = response_data.get('device_code')
            self.interval = response_data.get('interval')
            verification_url = response_data.get('verification_uri_complete')
            user_code = response_data.get('user_code')
            self.display_device_auth_flow(
                user_code=user_code, verification_url=verification_url)
            self.handle_auth_flow()

        else:
            raise Exception("Failed to initiate device flow auth.")

    def display_device_auth_flow(self, user_code: str, verification_url: str) -> None:
        """Displays the current device auth flow challenge - first by trying to
        open a browser window - if this fails then prints suggestion to stdout
        to try using the URL manually.

        Parameters
        ----------
        user_code : str
            The user code
        verification_url : str
            The url which embeds challenge code
        """
        print(f"Verification URL: {verification_url}")
        print(f"User Code: {user_code}")
        try:
            webbrowser.open(verification_url)
        except Exception:
            print("Tried to open web-browser but failed. Please visit URL above.")

    def handle_auth_flow(self) -> None:
        """Handles the device authorisation flow by constantly polling the token endpoint until a token
        is received, an error is received or a timeout occurs.
        """

        device_grant_type = "urn:ietf:params:oauth:grant-type:device_code"

        data = {
            "grant_type": device_grant_type,
            "device_code": self.device_code,
            "client_id": self.client_id,
            "scope": " ".join(self.scopes)
        }

        # Setup success criteria
        succeeded = False
        timed_out = False
        misc_fail = False

        # start time
        response_data: Optional[Dict[str, Any]] = None

        # Poll for success
        while not succeeded and not timed_out and not misc_fail:
            response = requests.post(self.token_endpoint, data=data)
            response_data = response.json()

            if response_data is None:
                misc_fail = True
                self.logger.info("No data received in the response.")

            elif response_data.get('error'):
                error = response_data['error']
                if error != 'authorization_pending':
                    misc_fail = True
                # Wait appropriate OAuth poll interval
                time.sleep(self.interval)
            else:

                # Successful as there was no error at the endpoint
                # We will produce a token object here.

                access_token = response_data.get("access_token")
                refresh_token = response_data.get("refresh_token")

                if not access_token:
                    misc_fail = True
                    self.logger.info("Missing or invalid access token.")
                    continue  # Skip this iteration, as we were not able to obtain a successful token

                if not refresh_token:
                    misc_fail = True
                    self.logger.info("Missing or invalid refresh token.")
                    continue  # Skip this iteration, as we were not able to obtain a successful token

                self.tokens = Tokens(
                    access_token=access_token,
                    refresh_token=refresh_token
                )

                # Save the tokens into '.token.json'
                self.save_tokens(self.tokens)
                succeeded = True

        if not succeeded:
            if response_data and "error" in response_data:
                self.logger.info(f"Failed due to {response_data['error']}")

            else:
                self.logger.info(
                    f"Failed with unknown error, failed to find error message.")


class OfflineFlow(AuthManager):
    # The keycloak endpoint to target for tokens
    keycloak_endpoint: str
    # The offline token provided by user
    offline_token: str
    # The client ID to target for auth
    client_id: str
    # The token endpoint to use
    token_endpoint: str

    scopes: list

    public_key: str

    def __init__(self, config: Config, client_id: str, offline_token: Optional[str] = None, offline_token_file: Optional[str] = None, log_level: Optional[LogType] = None) -> None:
        f"""Create and generate an OfflineFlow object. Instatiate from provided offline token, or attempt to read
        one from file and generate the access token. Can provide the offline token directly, a file for it stored as plain text.

        Parameters
        ----------
        keycloak_endpoint : str
            The keycloak endpoint to use. E.g., https://auth.example.org/auth/realms/my_realm",
        client_id : str
            The client to target for auth. E.g., landing-portal-ui
        offline_token : Optional[str], optional
            The offline token to use for generating access tokens from. If not provided, defaults to None and init will try use offline_token_file to read an offline_token.
        offline_token_file : Optional[str], optional
            The file name to read the offline token from, where it is stored as plain text. Be sure to add this file to your .gitignore if using this parameter.

        Raises
        ------
        Exception
            Fails to retrive public key from keycloak endpoint.
        Exception
            Fails to validate new tokens generated from using the supplied offline token (Only if offline token is provided)
        Exception
            Fails to validate refreshed tokens from using the offline token in the default file (Only if loading from file)
        """
        # construct parent class and include log level
        super().__init__(log_level=log_level)

        self.keycloak_endpoint = config.keycloak_endpoint
        self.client_id = client_id
        self.scopes: list = []

        self.token_endpoint = f'{self.keycloak_endpoint}/protocol/openid-connect/token'

        try:
            # First thing to do here is obtain the keycloak public key.
            self.public_key = retrieve_keycloak_public_key(
                logger=self.logger,
                keycloak_endpoint=self.keycloak_endpoint,
            )
        except Exception as e:
            raise Exception(
                "Failed to retrieve the Keycloak public key, authentication cannot proceed.") from e

        if not offline_token and not offline_token_file:
            err_msg = "Please provide a value or offline_token or offline_token_file."
            self.logger.error(err_msg)
            raise ValueError(err_msg)

        if offline_token:
            self.logger.info(
                "Offline token provided, attempting to generate tokens from it.")
            self.offline_token = offline_token
        elif offline_token_file:
            self.logger.info(
                "Offline token file provided, attempting to generate tokens from it.")
            self.offline_token = self.load_offline_token(offline_token_file)
        else:
            err_msg = "Please provide a value or offline_token or offline_token_file."
            self.logger.error(err_msg)
            raise ValueError(err_msg)

        # Ok, got an offline token, now generate an temporary access token from it
        try:
            self.get_access_token_from_offline_token()
        except Exception as e:
            err_msg = "Failed to validate new tokens generated from offline token file."
            self.logger.error(err_msg)
            raise Exception(err_msg) from e

    def get_token(self) -> str:
        """
        IMPLEMENTS BASE METHOD

        Uses the current token - validates it, 
        refreshes if necessary, and returns the valid token
        ready to be used.

        Returns
        -------
        str
            The access token

        Raises
        ------
        Exception
            Raises exception if tokens/public_key are not setup - make sure 
            that the object is instantiated properly before calling this function.
        Exception
            If the token is invalid and cannot be refreshed.
        Exception
            If the token validation still fails after re-conducting the device flow.
        """

        if self.tokens is None or self.public_key is None:
            raise Exception(
                "Cannot generate token without access token or public key.")

        try:
            # Attempt to validate the current token.
            if validate_access_token(
                    logger=self.logger,
                access_token=self.tokens.access_token,
                public_key=self.public_key
            ):
                return self.tokens.access_token
            else:
                self.logger.info("Token was invalid. Attempting Refresh")
                # refresh with refresh token and attempt re validation
                self.get_access_token_from_offline_token()
                if validate_access_token(
                    logger=self.logger,
                    access_token=self.tokens.access_token,
                    public_key=self.public_key
                ):
                    return self.tokens.access_token
                # still here, error
                err_msg = "Failed to produce a valid access token from the offline token."
                self.logger.error(err_msg)
                raise Exception(
                    err_msg)
        except Exception as e:
            err_msg = "Failed to refresh token."
            self.logger.error(err_msg)
            raise Exception(err_msg) from e

    def force_refresh(self) -> None:
        """
        IMPLEMENTS BASE METHOD 

        A method to reset the current authentication state.

        Since the offline flow has no cached state - this just forces a refresh
        token request to be made.
        """

        self.get_access_token_from_offline_token()

    def get_access_token_from_offline_token(self) -> None:
        tokens = keycloak_refresh_token_request(
            logger=self.logger,
            client_id=self.client_id,
            token_endpoint=self.token_endpoint,
            scopes=self.scopes,
            refresh_token=self.offline_token
        )

        access_token = tokens.get('access_token')
        if not access_token:
            err_msg = "Failed to geneate access token. Returned access token is None."
            self.logger.error(err_msg)
            raise ValueError(err_msg)

        self.tokens = AccessToken(
            access_token=access_token
        )

    def load_offline_token(self, file_name: str) -> str:
        """Loads the offline token from the provided file.

        Parameters
        ----------
        file_name : str
            The file name to load the offline token from.

        Returns
        -------
        str
            The offline token read from the file.

        Raises
        ------
        Exception
            If the file does not exist or if the file is empty.
        """

        if not os.path.exists(file_name):
            err_msg = f"Offline token file named {file_name} does not exist."
            self.logger.error(err_msg)
            raise Exception(err_msg)

        with open(file_name, 'r') as f:
            offline_token = f.read().strip()

        if not offline_token or offline_token == "":
            err_msg = f"File {file_name} is empty."
            self.logger.error(err_msg)
            raise Exception(err_msg)

        return offline_token



'''
Created Date: Monday June 17th 2024 +1000
Author: Peter Baker
-----
Last Modified: Monday June 17th 2024 4:45:39 pm +1000
Modified By: Peter Baker
-----
Description: General interfaces defined to help with typing the client library/configuring
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

from pydantic import BaseModel


class HealthCheckResponse(BaseModel):
    message: str

class AsyncAwaitSettings(BaseModel):
    # polling interval in seconds (defaults to 2 seconds)
    job_polling_interval : int = 2
        
    # how long do we wait for the entry to be present in table? (seconds)
    job_async_queue_delay_polling_timeout = 20  # 20 seconds

    # how long do we wait for it to leave pending? (seconds)
    job_async_pending_polling_timeout = 120  # 2 minutes

    # how long do we wait for it to become in progress? (seconds)
    job_async_in_progress_polling_timeout = 180  # 3 minutes
   
DEFAULT_AWAIT_SETTINGS = AsyncAwaitSettings()


'''
Created Date: Monday June 17th 2024 +1000
Author: Peter Baker
-----
Last Modified: Monday June 17th 2024 4:45:39 pm +1000
Modified By: Peter Baker
-----
Description: Job API L3 Module. Includes Job Admin sub module.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

from provenaclient.auth.manager import AuthManager
from provenaclient.utils.config import Config
from provenaclient.clients import JobAPIClient
from provenaclient.utils.exceptions import *
from provenaclient.modules.module_helpers import *
from provenaclient.models import HealthCheckResponse, AsyncAwaitSettings, DEFAULT_AWAIT_SETTINGS
from provenaclient.utils.async_job_helpers import wait_for_full_lifecycle, wait_for_full_successful_lifecycle
from ProvenaInterfaces.AsyncJobAPI import *
from typing import List, AsyncGenerator

# L3 interface.


class JobAdminSubService(ModuleService):
    _job_api_client: JobAPIClient

    def __init__(self, auth: AuthManager, config: Config, job_api_client: JobAPIClient) -> None:
        """Initialises a new job admin sub-service object, which sits between the user and the job-service api operations.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance. 
        job_api_client : JobAPIClient
            This client interacts with the Job API
        """
        self._auth = auth
        self._config = config

        # Clients related to the job-api scoped as private.
        self._job_api_client = job_api_client

    async def launch_job(self, request: AdminLaunchJobRequest) -> AdminLaunchJobResponse:
        """
        Launches a new job.

        Args:
            request (AdminLaunchJobRequest): Specified job parameters

        Returns:
            AdminLaunchJobResponse: The response
        """
        return await self._job_api_client.admin.launch_job(request=request)

    async def get_job(self, session_id: str) -> AdminGetJobResponse:
        """
        Fetches a job (any job since admin)

        Args:
            session_id (str): The session ID of job to fetch

        Returns:
            AdminGetJobResponse: The response
        """
        return await self._job_api_client.admin.get_job(session_id=session_id)

    async def list_jobs(self, list_jobs_request: AdminListJobsRequest) -> AdminListJobsResponse:
        """
        Lists all jobs.

        Args:
            list_jobs_request (AdminListJobsRequest): The request including pagination information.

        Returns:
            AdminListJobsResponse: The list of jobs response.
        """
        return await self._job_api_client.admin.list_jobs(list_jobs_request=list_jobs_request)
    
    async def list_all_jobs(self, list_jobs_request: AdminListJobsRequest, limit: Optional[int] = None) -> List[JobStatusTable]:
        """
        Lists all jobs for the given user.

        Will automatically paginate until list is exhausted.

        Args:
            list_jobs_request (AdminListJobsRequest): The request including details
            limit (Optional[int]): Total record limit to enforce, if any

        Returns:
            ListJobsResponse: The list of jobs
        """
        all_jobs: List[JobStatusTable] = []
        count = 0

        # paginate until limit hit
        while True:
            new_list = await self._job_api_client.admin.list_jobs(list_jobs_request=list_jobs_request)
            count += len(new_list.jobs)
            all_jobs.extend(new_list.jobs)

            if limit is not None and count >= limit:
                break

            if new_list.pagination_key is None:
                break

            list_jobs_request.pagination_key = new_list.pagination_key

        return all_jobs

    async def for_all_jobs(self, list_jobs_request: AdminListJobsRequest, limit: Optional[int] = None) -> AsyncGenerator[JobStatusTable, None]:
        """
        Lists all jobs for the given user.

        Returns lazy generator for use in for loops.

        Will automatically paginate until list is exhausted.

        Args:
            list_jobs_request (AdminListJobsRequest): The request including details
            limit (Optional[int]): Total record limit to enforce, if any

        Returns:
            ListJobsResponse: The list of jobs
        """
        count = 0

        # paginate until limit hit
        while True:
            new_list = await self._job_api_client.admin.list_jobs(list_jobs_request=list_jobs_request)
            count += len(new_list.jobs)

            for job in new_list.jobs:
                yield job

            if limit is not None and count >= limit:
                break

            if new_list.pagination_key is None:
                break

            list_jobs_request.pagination_key = new_list.pagination_key
    

    async def list_job_batch(self, list_request: AdminListByBatchRequest) -> AdminListByBatchResponse:
        """
        List jobs by batch ID, returning a list of jobs in the batch.

        Args:
            list_request (AdminListByBatchRequest): The request including batch ID

        Returns:
            AdminListByBatchResponse: The list of jobs in the batch
        """
        return await self._job_api_client.admin.list_jobs_in_batch(list_request=list_request)

    async def list_all_jobs_in_batch(self, list_request: AdminListByBatchRequest, limit: Optional[int] = None) -> List[JobStatusTable]:
        """
        Lists all jobs for the given user. Will automatically paginate all
        entries to exhaust list 

        NOTE this could return more than limit - but figure it may as well
        return data fetched for efficiency reasons

        Args:
            list_jobs_request (AdminListJobsRequest): The request including details
            limit (Optional[int]): Total record limit to enforce, if any

        Returns:
            ListJobsResponse: The list of jobs
        """
        all_jobs: List[JobStatusTable] = []
        count = 0

        # paginate until limit hit
        while True:
            new_list = await self._job_api_client.admin.list_jobs_in_batch(list_request=list_request)
            count += len(new_list.jobs)
            all_jobs.extend(new_list.jobs)

            if limit is not None and count >= limit:
                break

            if new_list.pagination_key is None:
                break

            list_request.pagination_key = new_list.pagination_key

        return all_jobs

    async def for_all_jobs_in_batch(self, list_request: AdminListByBatchRequest, limit: Optional[int] = None) -> AsyncGenerator[JobStatusTable, None]:
        """
        Lists all jobs for the given user. Will automatically paginate all
        entries to exhaust list

        NOTE this could return more than limit - but figure it may as well
        return data fetched for efficiency reasons

        Args:
            list_jobs_request (AdminListJobsRequest): The request including details
            limit (Optional[int]): Total record limit to enforce, if any

        Returns:
            ListJobsResponse: The list of jobs
        """
        count = 0

        # paginate until limit hit
        while True:
            new_list = await self._job_api_client.admin.list_jobs_in_batch(list_request=list_request)
            count += len(new_list.jobs)

            for job in new_list.jobs:
                yield job

            if limit is not None and count >= limit:
                break

            if new_list.pagination_key is None:
                break

            list_request.pagination_key = new_list.pagination_key

class JobService(ModuleService):
    _job_api_client: JobAPIClient

    admin: JobAdminSubService

    def __init__(self, auth: AuthManager, config: Config, job_api_client: JobAPIClient) -> None:
        """Initialises a new job-service object, which sits between the user and the job-service api operations.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance. 
        job_api_client : JobAPIClient
            This client interacts with the Job API
        """
        self._auth = auth
        self._config = config

        # Clients related to the job-api scoped as private.
        self._job_api_client = job_api_client

        # Sub modules
        self.admin = JobAdminSubService(
            auth=auth, config=config, job_api_client=job_api_client
        )

    async def get_health_check(self) -> HealthCheckResponse:
        """
        Health check the API

        Returns:
            HealthCheckResponse: Response
        """
        return await self._job_api_client.get_health_check()

    async def fetch_job(self, session_id: str) -> GetJobResponse:
        """
        Fetches a job by session id

        Args:
            session_id (str): The session ID

        Returns:
            GetJobResponse: The job fetched
        """
        return await self._job_api_client.fetch_job(session_id=session_id)

    async def list_jobs(self, list_jobs_request: ListJobsRequest) -> ListJobsResponse:
        """
        Lists all jobs for the given user

        Can return pagination if page size exceeded

        Args:
            list_jobs_request (ListJobsRequest): The request including details

        Returns:
            ListJobsResponse: The list of jobs
        """
        return await self._job_api_client.list_jobs(list_jobs_request=list_jobs_request)

    async def list_all_jobs(self, list_jobs_request: ListJobsRequest, limit: Optional[int] = None) -> List[JobStatusTable]:
        """
        Lists all jobs for the given user.

        Will automatically paginate until list is exhausted.

        Args:
            list_jobs_request (ListJobsRequest): The request including details
            limit (Optional[int]): Total record limit to enforce, if any

        Returns:
            ListJobsResponse: The list of jobs
        """
        all_jobs: List[JobStatusTable] = []
        count = 0

        # paginate until limit hit
        while True:
            new_list = await self._job_api_client.list_jobs(list_jobs_request=list_jobs_request)
            count += len(new_list.jobs)
            all_jobs.extend(new_list.jobs)

            if limit is not None and count >= limit:
                break

            if new_list.pagination_key is None:
                break

            list_jobs_request.pagination_key = new_list.pagination_key

        return all_jobs

    async def for_all_jobs(self, list_jobs_request: ListJobsRequest, limit: Optional[int] = None) -> AsyncGenerator[JobStatusTable, None]:
        """
        Lists all jobs for the given user.

        Returns lazy generator for use in for loops.

        Will automatically paginate until list is exhausted.

        Args:
            list_jobs_request (ListJobsRequest): The request including details
            limit (Optional[int]): Total record limit to enforce, if any

        Returns:
            ListJobsResponse: The list of jobs
        """
        count = 0

        # paginate until limit hit
        while True:
            new_list = await self._job_api_client.list_jobs(list_jobs_request=list_jobs_request)
            count += len(new_list.jobs)

            for job in new_list.jobs:
                yield job

            if limit is not None and count >= limit:
                break

            if new_list.pagination_key is None:
                break

            list_jobs_request.pagination_key = new_list.pagination_key

    async def list_jobs_in_batch(self, list_request: ListByBatchRequest) -> ListByBatchResponse:
        """
        Gets all jobs within a batch.

        Can return pagination if page size exceeded

        Args:
            list_request (ListByBatchRequest): The request including batch ID

        Returns:
            ListByBatchResponse: The response including list of jobs
        """
        return await self._job_api_client.list_jobs_in_batch(list_request=list_request)

    async def list_all_jobs_in_batch(self, list_request: ListByBatchRequest, limit: Optional[int] = None) -> List[JobStatusTable]:
        """
        Lists all jobs for the given user. Will automatically paginate all
        entries to exhaust list 

        NOTE this could return more than limit - but figure it may as well
        return data fetched for efficiency reasons

        Args:
            list_jobs_request (ListJobsRequest): The request including details
            limit (Optional[int]): Total record limit to enforce, if any

        Returns:
            ListJobsResponse: The list of jobs
        """
        all_jobs: List[JobStatusTable] = []
        count = 0

        # paginate until limit hit
        while True:
            new_list = await self._job_api_client.list_jobs_in_batch(list_request=list_request)
            count += len(new_list.jobs)
            all_jobs.extend(new_list.jobs)

            if limit is not None and count >= limit:
                break

            if new_list.pagination_key is None:
                break

            list_request.pagination_key = new_list.pagination_key

        return all_jobs

    async def for_all_jobs_in_batch(self, list_request: ListByBatchRequest, limit: Optional[int] = None) -> AsyncGenerator[JobStatusTable, None]:
        """
        Lists all jobs for the given user. Will automatically paginate all
        entries to exhaust list

        NOTE this could return more than limit - but figure it may as well
        return data fetched for efficiency reasons

        Args:
            list_jobs_request (ListJobsRequest): The request including details
            limit (Optional[int]): Total record limit to enforce, if any

        Returns:
            ListJobsResponse: The list of jobs
        """
        count = 0

        # paginate until limit hit
        while True:
            new_list = await self._job_api_client.list_jobs_in_batch(list_request=list_request)
            count += len(new_list.jobs)

            for job in new_list.jobs:
                yield job

            if limit is not None and count >= limit:
                break

            if new_list.pagination_key is None:
                break

            list_request.pagination_key = new_list.pagination_key

    async def await_job_completion(self, session_id: str, settings: AsyncAwaitSettings = DEFAULT_AWAIT_SETTINGS) -> JobStatusTable:
        """

        Awaits completion of a given job then provides the job info.

        Completion is defined as a job status which is not pending or in progress.

        Args:
            session_id (str): The ID of the job to monitor and await completion.

        Returns:
            JobStatusTable: The entry at the latest point.
        """

        return await wait_for_full_lifecycle(
            session_id=session_id,
            client=self._job_api_client,
            settings=settings
        )

    async def await_successful_job_completion(self, session_id: str, settings: AsyncAwaitSettings = DEFAULT_AWAIT_SETTINGS) -> JobStatusTable:
        """

        Awaits successful completion of a given job then provides the job info.

        Completion is defined as a job status which is not pending or in progress.

        Args:
            session_id (str): The ID of the job to monitor and await completion.

        Returns:
            JobStatusTable: The entry at the latest point.
        """

        return await wait_for_full_successful_lifecycle(
            session_id=session_id,
            client=self._job_api_client,
            settings=settings
        )



'''
Created Date: Monday June 17th 2024 +1000
Author: Peter Baker
-----
Last Modified: Monday June 17th 2024 4:45:39 pm +1000
Modified By: Peter Baker
-----
Description: Auth Manager interface which defines key methods for authorising API requests in the provena client
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

from abc import ABC, abstractmethod
from provenaclient.auth.helpers import HttpxBearerAuth
from typing import Optional, Literal
from enum import Enum
import logging


class Log(Enum):
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR

# Type alias so we can type the function properly using high level types
LogType = Literal[Log.DEBUG, Log.INFO, Log.WARNING, Log.ERROR]

DEFAULT_LOG_LEVEL = Log.ERROR


class AuthManager(ABC):
    # This must
    logger: logging.Logger

    def __init__(self, log_level: Optional[LogType] = None) -> None:
        # Create a logger
        self.logger = logging.getLogger('auth-logger')

        # LOGGING LEVEL
        self.logger.setLevel(
            log_level.value if log_level else DEFAULT_LOG_LEVEL.value)

        # Create a console handler and set its log level
        console_handler = logging.StreamHandler()
        console_handler.setLevel(log_level.value if log_level else DEFAULT_LOG_LEVEL.value)

        # Create a formatter and set it for the console handler
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        console_handler.setFormatter(formatter)

        # Add the handler to the logger
        self.logger.addHandler(console_handler)

    @abstractmethod
    def get_token(self) -> str:
        """Get token information and other metadata."""
        pass

    @abstractmethod
    def force_refresh(self) -> None:
        """ Force refresh the current token"""
        pass

    def get_auth(self) -> HttpxBearerAuth:
        """A helper function which produces a BearerAuth object for use
        in the httpx library. For example: 

        manager = DeviceFlow(...)
        auth = manager.get_auth 
        httpx.post(..., auth=auth)

        Returns
        -------
        BearerAuth
            The httpx auth object.

        Raises
        ------
        Exception
            Raises exception if tokens/public_key are not setup - make sure 
            that the object is instantiated properly before calling this function.
        Exception
            If the token is invalid and cannot be refreshed.
        Exception
            If the token validation still fails after re-conducting the device flow.
        """

        token = self.get_token()
        return HttpxBearerAuth(token=token)



'''
Created Date: Monday June 17th 2024 +1000
Author: Peter Baker
-----
Last Modified: Monday June 17th 2024 5:00:47 pm +1000
Modified By: Peter Baker
-----
Description: A HTTP client which wraps HTTPx async so that we can swap out a different http library later if needed.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

from typing import Any, List, Optional, Union
import httpx
from provenaclient.auth.helpers import HttpxBearerAuth
from provenaclient.utils.helpers import JsonData

# 60s timeout for connecting, and a 10s timeout elsewhere.
timeout = httpx.Timeout(timeout=10.0, connect=60.0)

# L1 interface.


class HttpClient:

    """This class only contains static methods as it acts as an HTTP client and provides a layer over these static methods 
    and makes them easily identifiable within the codebase.
    """

    @staticmethod
    async def make_get_request(url: str, params: Optional[dict[str, Any]] = None, auth: Optional[HttpxBearerAuth] = None, headers: Optional[dict[str, Any]] = None) -> httpx.Response:
        """ Makes an asynchronous HTTP GET request to the specified URL using the provided parameters, authentication, and headers.

        Parameters
        ----------
        url : str
            The URL to which the GET request will be sent.
        params : Optional[dict[str, Any]], optional
            A dictionary of the query parameters to be included in the GET request, by default None.
        auth : Optional[HttpxBearerAuth], optional
            Authentication object (httpx bearer token only), to be included in the request headers, by default None.
        headers : Optional[dict[str,Any]], optional
            A dictionary having additional HTTP headers to send with the GET request, by default None.

        Returns
        -------
        httpx.Response
            The response from the server as an httpx.Response object.
        """
        async with httpx.AsyncClient(timeout=timeout) as client:
            response = await client.get(url, params=params, headers=headers, auth=auth)
            return response

    @staticmethod
    async def make_delete_request(url: str,  auth: HttpxBearerAuth, params: Optional[dict[str, Any]] = None, headers: Optional[dict[str, Any]] = None) -> httpx.Response:
        """ Makes an asynchronous HTTP DELETE request to the specified URL using the provided parameters, authentication, and headers.

        Parameters
        ----------
        url : str
            The URL to which the DELETE request will be sent.
        params : Optional[dict[str, Any]], optional
            A dictionary of the query parameters to be included in the GET request, by default None.
        auth : Optional[HttpxBearerAuth], optional
            Authentication object (httpx bearer token only), to be included in the request headers, by default None.
        headers : Optional[dict[str,Any]], optional
            A dictionary having additional HTTP headers to send with the GET request, by default None.

        Returns
        -------
        httpx.Response
            The response from the server as an httpx.Response object.
        """
        async with httpx.AsyncClient(timeout=timeout) as client:
            response = await client.delete(url, params=params, headers=headers, auth=auth)
            return response

    @staticmethod
    async def make_post_request(url: str, auth: HttpxBearerAuth, params: Optional[dict[str, Any]] = None, data: Union[Optional[dict[str, Any]], Optional[List[dict[str,Any]]]] = None, files: Optional[dict[str, tuple[str, bytes, str]]] = None, headers: Optional[dict[str, Any]] = None) -> httpx.Response:
        """ Makes an asynchronous HTTP POST request to the specified URL with the provided data, authentication, and headers.

        Parameters
        ----------
        url : str
            The URL to which the GET request will be sent.
        params : Optional[dict[str, Any]], optional
            A dictionary of the query parameters to be included in the GET request, by default None.
        auth : HttpxBearerAuth
            Authentication object (e.g., bearer token), which is required for the POST request.
        data : Optional[dict[str,Any]], optional
            A dictionary of the data to be sent in the body of the POST request, by default None.
        files: Optional[dict[str, tuple[str, bytes, str]]], optional
            A files request object containing the file content and the media type.
        headers : Optional[dict[str, Any]], optional
            A dictionary representing additional HTTP headers to send with the POST request, by default None.

        Returns
        -------
        httpx.Response
            The response from the server as an httpx.Response object.
        """
        async with httpx.AsyncClient(timeout=timeout) as client:
            response = await client.post(url, params=params, json=data, headers=headers, files=files, auth=auth)
            return response

    @staticmethod
    async def make_put_request(url: str, auth: HttpxBearerAuth, params: Optional[dict[str, Any]] = None, data: Optional[JsonData] = None, headers: Optional[dict[str, Any]] = None) -> httpx.Response:
        """ Makes an asynchronous HTTP put request to the specified URL with the provided data, authentication, and headers.

        Parameters
        ----------
        url : str
            The URL to which the GET request will be sent.
        params : Optional[dict[str, Any]], optional
            A dictionary of the query parameters to be included in the GET request, by default None.
        auth : HttpxBearerAuth
            Authentication object (e.g., bearer token), which is required for the put request.
        data : Optional[dict[str,Any]], optional
            A dictionary of the data to be sent in the body of the put request, by default None.
        headers : Optional[dict[str, Any]], optional
            A dictionary representing additional HTTP headers to send with the put request, by default None.

        Returns
        -------
        httpx.Response
            The response from the server as an httpx.Response object.
        """
        async with httpx.AsyncClient(timeout=timeout) as client:
            response = await client.put(url, params=params, json=data, headers=headers, auth=auth)
            return response



'''
Created Date: Monday June 17th 2024 +1000
Author: Peter Baker
-----
Last Modified: Monday June 17th 2024 4:45:39 pm +1000
Modified By: Peter Baker
-----
Description: Helper methods for clients namely a standard interface ClientService along with reusable patterns to interact with the HTTPX L1 Layer
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------

18-06-2024 | Peter Baker | Just noting I think we could streamline the helper functions a bit
'''

from abc import ABC
from provenaclient.auth import AuthManager
from provenaclient.utils.config import Config
from provenaclient.utils.helpers import *
from provenaclient.utils.http_client import HttpClient
from typing import Dict, Mapping, Optional
from provenaclient.utils.exceptions import CustomTimeoutException


class ClientService(ABC):
    """
    This class interface just captures that the client has an instantiated auth
    manager which allows for helper functions abstracted for L2 clients.
    """
    _auth: AuthManager
    _config: Config


async def parsed_get_request_with_status(client: ClientService, params: Optional[Mapping[str, Optional[ParamTypes]]], url: str, error_message: str, model: Type[BaseModelType]) -> BaseModelType:
    """

    High level helper function which 

    - gets the auth
    - builds the filtered param list
    - makes get request
    - checks status/http codes
    - parses model etc

    Returns the parsed validated, safe to use 200OK model result

    Args:
        client (ClientService): The client being used. Relies on client interface.
        params (Optional[Mapping[str, Optional[ParamTypes]]]): The params if any
        url (str): The url to make GET request to
        error_message (str): The error message to embed in other exceptions
        model (Type[BaseModelType]): Model to parse for response JSON

    Raises:
        e: Exception depending on error

    Returns:
        BaseModelType: The specified parsed model
    """
    # Prepare and setup the API request.
    get_auth = client._auth.get_auth  # Get bearer auth
    filtered_params = build_params_exclude_none(params if params else {})

    try:
        response = await HttpClient.make_get_request(url=url, params=filtered_params, auth=get_auth())
        data = handle_response_with_status(
            response=response,
            model=model,
            error_message=error_message
        )

    except BaseException as e:
        raise e
    except Exception as e:
        raise Exception(
            f"{error_message} Exception: {e}") from e

    return data


async def parsed_get_request(client: ClientService, params: Optional[Mapping[str, Optional[ParamTypes]]], url: str, error_message: str, model: Type[BaseModelType]) -> BaseModelType:
    """

    High level helper function which 

    - gets the auth
    - builds the filtered param list
    - makes get request
    - checks http codes
    - parses model etc

    Returns the parsed validated, safe to use 200OK model result

    Args:
        client (ClientService): The client being used. Relies on client interface.
        params (Optional[Mapping[str, Optional[ParamTypes]]]): The params if any
        url (str): The url to make GET request to
        error_message (str): The error message to embed in other exceptions
        model (Type[BaseModelType]): Model to parse for response JSON

    Raises:
        e: Exception depending on error

    Returns:
        BaseModelType: The specified parsed model
    """
    # Prepare and setup the API request.
    get_auth = client._auth.get_auth  # Get bearer auth
    filtered_params = build_params_exclude_none(params if params else {})

    try:
        response = await HttpClient.make_get_request(url=url, params=filtered_params, auth=get_auth())
        data = handle_response_non_status(
            response=response,
            model=model,
            error_message=error_message
        )

    except BaseException as e:
        raise e
    except Exception as e:
        raise Exception(
            f"{error_message} Exception: {e}") from e

    return data


async def parsed_post_request(client: ClientService, params: Optional[Mapping[str, Optional[ParamTypes]]], json_body: Optional[JsonData], url: str, error_message: str, model: Type[BaseModelType]) -> BaseModelType:
    """

    High level helper function which 

    - gets the auth
    - builds the filtered param list
    - makes POST request
    - checks http codes
    - parses model etc

    Returns the parsed validated, safe to use 200OK model result

    Args:
        client (ClientService): The client being used. Relies on client interface.
        params (Optional[Mapping[str, Optional[ParamTypes]]]): The params if any
        url (str): The url to make POST request to
        error_message (str): The error message to embed in other exceptions
        model (Type[BaseModelType]): Model to parse for response JSON
        json_body: Optional[JsonData]: JSON data to post if any

    Raises:
        e: Exception depending on error

    Returns:
        BaseModelType: The specified parsed model
    """
    # Prepare and setup the API request.
    get_auth = client._auth.get_auth  # Get bearer auth
    filtered_params = build_params_exclude_none(params if params else {})

    try:
        response = await HttpClient.make_post_request(url=url, data=json_body, params=filtered_params, auth=get_auth())
        data = handle_response_non_status(
            response=response,
            model=model,
            error_message=error_message
        )

    except BaseException as e:
        raise e
    except Exception as e:
        raise Exception(
            f"{error_message} Exception: {e}") from e

    return data

async def parsed_post_request_with_status(client: ClientService, params: Optional[Mapping[str, Optional[ParamTypes]]], json_body: Optional[JsonData], url: str, error_message: str, model: Type[BaseModelType], files: Optional[HttpxFileUpload] = None) -> BaseModelType:
    """

    High level helper function which 

    - gets the auth
    - builds the filtered param list
    - makes POST request
    - checks http/status codes
    - parses model etc

    Returns the parsed validated, safe to use 200OK model result

    Args:
        client (ClientService): The client being used. Relies on client interface.
        params (Optional[Mapping[str, Optional[ParamTypes]]]): The params if any
        url (str): The url to make POST request to
        error_message (str): The error message to embed in other exceptions
        model (Type[BaseModelType]): Model to parse for response JSON
        json_body: Optional[JsonData]: JSON data to post if any
        files: Optional[HttpxFileUpload]: A dictionary representing file(s) to be uploaded with the
               request. Each key in the dictionary is the name of the form field for the file according,
               to API specifications. For Provena it's "csv_file" and and the value 
               is a tuple of (filename, filedata, MIME type or media type).

    Raises:
        e: Exception depending on error

    Returns:
        BaseModelType: The specified parsed model
    """
    # Prepare and setup the API request.
    get_auth = client._auth.get_auth  # Get bearer auth
    filtered_params = build_params_exclude_none(params if params else {})

    try:
        response = await HttpClient.make_post_request(url=url, data=json_body, params=filtered_params, files = files, auth=get_auth())
        data = handle_response_with_status(
            response=response,
            model=model,
            error_message=error_message
        )

    except BaseException as e:
        raise e
    except Exception as e:
        raise Exception(
            f"{error_message} Exception: {e}") from e

    return data


async def parsed_delete_request_with_status(client: ClientService, params: Optional[Mapping[str, Optional[ParamTypes]]], url: str, error_message: str, model: Type[BaseModelType]) -> BaseModelType:
    """

    High level helper function which 

    - gets the auth
    - builds the filtered param list
    - makes DELETE request
    - checks http/status codes
    - parses model etc

    Returns the parsed validated, safe to use 200OK model result

    Args:
        client (ClientService): The client being used. Relies on client interface.
        params (Optional[Mapping[str, Optional[ParamTypes]]]): The params if any
        url (str): The url to make POST request to
        error_message (str): The error message to embed in other exceptions
        model (Type[BaseModelType]): Model to parse for response JSON
        json_body: Optional[JsonData]: JSON data to post if any

    Raises:
        e: Exception depending on error

    Returns:
        BaseModelType: The specified parsed model
    """
    # Prepare and setup the API request.
    get_auth = client._auth.get_auth  # Get bearer auth
    filtered_params = build_params_exclude_none(params if params else {})

    try:
        response = await HttpClient.make_delete_request(url=url, params=filtered_params, auth=get_auth())
        data = handle_response_with_status(
            response=response,
            model=model,
            error_message=error_message
        )

    except BaseException as e:
        raise e
    except Exception as e:
        raise Exception(
            f"{error_message} Exception: {e}") from e

    return data


async def parsed_delete_request(client: ClientService, params: Optional[Mapping[str, Optional[ParamTypes]]], url: str, error_message: str, model: Type[BaseModelType]) -> BaseModelType:
    """

    High level helper function which 

    - gets the auth
    - builds the filtered param list
    - makes DELETE request
    - checks http/status codes
    - parses model etc

    Returns the parsed validated, safe to use 200OK model result

    Args:
        client (ClientService): The client being used. Relies on client interface.
        params (Optional[Mapping[str, Optional[ParamTypes]]]): The params if any
        url (str): The url to make POST request to
        error_message (str): The error message to embed in other exceptions
        model (Type[BaseModelType]): Model to parse for response JSON
        json_body: Optional[JsonData]: JSON data to post if any

    Raises:
        e: Exception depending on error

    Returns:
        BaseModelType: The specified parsed model
    """
    # Prepare and setup the API request.
    get_auth = client._auth.get_auth  # Get bearer auth
    filtered_params = build_params_exclude_none(params if params else {})

    try:
        response = await HttpClient.make_delete_request(url=url, params=filtered_params, auth=get_auth())
        data = handle_response_non_status(
            response=response,
            model=model,
            error_message=error_message
        )

    except BaseException as e:
        raise e
    except Exception as e:
        raise Exception(
            f"{error_message} Exception: {e}") from e

    return data


async def parsed_put_request(client: ClientService, params: Optional[Mapping[str, Optional[ParamTypes]]], json_body: Optional[JsonData], url: str, error_message: str, model: Type[BaseModelType]) -> BaseModelType:
    """

    High level helper function which 

    - gets the auth
    - builds the filtered param list
    - makes put request
    - checks http codes
    - parses model etc

    Returns the parsed validated, safe to use 200OK model result

    Args:
        client (ClientService): The client being used. Relies on client interface.
        params (Optional[Mapping[str, Optional[ParamTypes]]]): The params if any
        url (str): The url to make put request to
        error_message (str): The error message to embed in other exceptions
        model (Type[BaseModelType]): Model to parse for response JSON
        json_body: Optional[JsonData]: JSON data to put if any

    Raises:
        e: Exception depending on error

    Returns:
        BaseModelType: The specified parsed model
    """
    # Prepare and setup the API request.
    get_auth = client._auth.get_auth  # Get bearer auth
    filtered_params = build_params_exclude_none(params if params else {})

    try:
        response = await HttpClient.make_put_request(url=url, data=json_body, params=filtered_params, auth=get_auth())
        data = handle_response_non_status(
            response=response,
            model=model,
            error_message=error_message
        )

    except BaseException as e:
        raise e
    except Exception as e:
        raise Exception(
            f"{error_message} Exception: {e}") from e

    return data


async def parsed_put_request_with_status(client: ClientService, params: Optional[Mapping[str, Optional[ParamTypes]]], json_body: Optional[JsonData], url: str, error_message: str, model: Type[BaseModelType]) -> BaseModelType:
    """

    High level helper function which 

    - gets the auth
    - builds the filtered param list
    - makes put request
    - checks http/status codes
    - parses model etc

    Returns the parsed validated, safe to use 200OK model result

    Args:
        client (ClientService): The client being used. Relies on client interface.
        params (Optional[Mapping[str, Optional[ParamTypes]]]): The params if any
        url (str): The url to make put request to
        error_message (str): The error message to embed in other exceptions
        model (Type[BaseModelType]): Model to parse for response JSON
        json_body: Optional[JsonData]: JSON data to put if any

    Raises:
        e: Exception depending on error

    Returns:
        BaseModelType: The specified parsed model
    """
    # Prepare and setup the API request.
    get_auth = client._auth.get_auth  # Get bearer auth
    filtered_params = build_params_exclude_none(params if params else {})

    try:
        response = await HttpClient.make_put_request(url=url, data=json_body, params=filtered_params, auth=get_auth())
        data = handle_response_with_status(
            response=response,
            model=model,
            error_message=error_message
        )

    except BaseException as e:
        raise e
    except Exception as e:
        raise Exception(
            f"{error_message} Exception: {e}") from e

    return data



async def validated_get_request(client: ClientService, params: Optional[Mapping[str, Optional[ParamTypes]]], url: str, error_message: str) -> Response:
    """

    High level helper function which 

    - gets the auth
    - builds the filtered param list
    - makes get request
    - checks http codes

    This method does not do any base model parsing and only checks HTTP status codes.

    Args:
        client (ClientService): The client being used. Relies on client interface.
        params (Optional[Mapping[str, Optional[ParamTypes]]]): The params if any
        url (str): The url to make GET request to
        error_message (str): The error message to embed in other exceptions

    Raises:
        e: Exception depending on error

    Returns:
        None
    """
    # Prepare and setup the API request.
    get_auth = client._auth.get_auth  # Get bearer auth
    filtered_params = build_params_exclude_none(params if params else {})

    try:
        response = await HttpClient.make_get_request(url=url, params=filtered_params, auth=get_auth())
        
        handle_err_codes(
            response=response,
            error_message=error_message
        )

        return response

    except BaseException as e:
        raise e
    except Exception as e:
        raise Exception(
            f"{error_message} Exception: {e}") from e


async def parsed_post_request_none_return(client: ClientService, params: Optional[Mapping[str, Optional[ParamTypes]]], json_body: Optional[JsonData], url: str, error_message: str) -> None:
    """

    High level helper function which 

    - gets the auth
    - builds the filtered param list
    - makes POST request
    - checks http codes

    This method does not do any base model parsing and only checks HTTP status codes.

    Args:
        client (ClientService): The client being used. Relies on client interface.
        params (Optional[Mapping[str, Optional[ParamTypes]]]): The params if any
        url (str): The url to make POST request to
        error_message (str): The error message to embed in other exceptions
        json_body: Optional[JsonData]: JSON data to post if any

    Raises:
        e: Exception depending on error

    Returns:
        None
    """
    # Prepare and setup the API request.
    get_auth = client._auth.get_auth  # Get bearer auth
    filtered_params = build_params_exclude_none(params if params else {})

    try:
        response = await HttpClient.make_post_request(url=url, data=json_body, params=filtered_params, auth=get_auth())
        
        handle_err_codes(
            response=response,
            error_message=error_message
        )

    except BaseException as e:
        raise e
    except Exception as e:
        raise Exception(
            f"{error_message} Exception: {e}") from e
    

async def parsed_delete_request_non_return(client: ClientService, params: Optional[Mapping[str, Optional[ParamTypes]]], url: str, error_message: str) -> None:
    """

    High level helper function which 

    - gets the auth
    - builds the filtered param list
    - makes DELETE request
    - checks http/status codes

    This method does not do any base model parsing and only checks HTTP status codes.

    Args:
        client (ClientService): The client being used. Relies on client interface.
        params (Optional[Mapping[str, Optional[ParamTypes]]]): The params if any
        url (str): The url to make POST request to
        error_message (str): The error message to embed in other exceptions

    Raises:
        e: Exception depending on error

    Returns:
        None
    """
    # Prepare and setup the API request.
    get_auth = client._auth.get_auth  # Get bearer auth
    filtered_params = build_params_exclude_none(params if params else {})

    try:
        response = await HttpClient.make_delete_request(url=url, params=filtered_params, auth=get_auth())
        
        handle_err_codes(
            response=response,
            error_message=error_message
        )

    except BaseException as e:
        raise e
    except Exception as e:
        raise Exception(
            f"{error_message} Exception: {e}") from e



'''
Created Date: Thursday June 6th 2024 +1000
Author: Peter Baker
-----
Last Modified: Thursday June 6th 2024 1:39:55 pm +1000
Modified By: Peter Baker
-----
Description: Datastore L2 Client.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

from provenaclient.auth.manager import AuthManager
from provenaclient.utils.config import Config
from provenaclient.clients.client_helpers import *
from enum import Enum
from ProvenaInterfaces.DataStoreAPI import *
from provenaclient.models import HealthCheckResponse, RevertMetadata
from ProvenaInterfaces.RegistryModels import CollectionFormat
from ProvenaInterfaces.RegistryAPI import NoFilterSubtypeListRequest, VersionRequest, VersionResponse, DatasetListResponse
from provenaclient.utils.helpers import *


class DatastoreEndpoints(str, Enum):
    """An ENUM containing the datastore-api endpoints."""

    # Completed.
    GET_REGISTRY_ITEMS_FETCH_DATASET = "/registry/items/fetch-dataset"
    POST_REGISTER_MINT_DATASET = "/register/mint-dataset"
    GET_HEALTH_CHECK = "/"
    POST_METADATA_VALIDATE_METADATA = "/metadata/validate-metadata"
    GET_METADATA_DATASET_SCHEMA = "/metadata/dataset-schema"
    POST_REGISTER_UPDATE_METADATA = "/register/update-metadata"
    PUT_REGISTER_REVERT_METADATA = "/register/revert-metadata"
    POST_REGISTER_VERSION = "/register/version"
    POST_REGISTRY_ITEMS_LIST = "/registry/items/list"
    POST_REGISTRY_ITEMS_GENERATE_PRESIGNED_URL = "/registry/items/generate-presigned-url"
    POST_REGISTRY_CREDENTIALS_GENERATE_READ_ACCESS_CREDENTIALS = "/registry/credentials/generate-read-access-credentials"
    POST_REGISTRY_CREDENTIALS_GENERATE_WRITE_ACCESS_CREDENTIALS = "/registry/credentials/generate-write-access-credentials"
    DELETE_RELEASE_SYS_REVIEWERS_DELETE = "/release/sys-reviewers/delete"
    POST_RELEASE_SYS_REVIEWERS_ADD = "/release/sys-reviewers/add"
    GET_RELEASE_SYS_REVIEWERS_LIST = "/release/sys-reviewers/list"
    POST_RELEASE_APPROVAL_REQUEST = "/release/approval-request"
    PUT_RELEASE_ACTION_APPROVAL_REQUEST = "/release/action-approval-request"


    # Not completed.
    GET_CHECK_ACCESS_CHECK_GENERAL_ACCESS = "/check-access/check-general-access"
    GET_CHECK_ACCESS_CHECK_ADMIN_ACCESS = "/check-access/check-admin-access"
    GET_CHECK_ACCESS_CHECK_READ_ACCESS = "/check-access/check-read-access"
    GET_CHECK_ACCESS_CHECK_WRITE_ACCESS = "/check-access/check-write-access"
    GET_ADMIN_CONFIG = "/admin/config"
    GET_ADMIN_SENTRY_DEBUG = "/admin/sentry-debug"


# L2 interface.

class DatasetReviewSubClient(ClientService):

    def __init__(self, auth: AuthManager, config: Config) -> None:
        """Initialise the Datastore system reviewer sub client with authentication and configuration.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance.
        """
        self._auth = auth
        self._config = config

    def _build_endpoint(self, endpoint: DatastoreEndpoints) -> str:
        return self._config.auth_api_endpoint + endpoint.value
    
    
    async def delete_dataset_reviewer(self, reviewer_id: str) -> None:
        """Delete a reviewer existing within the datastore.

        Parameters
        ----------
        reviewer_id : str
            Id of an existing reviewer within the system.
        """

        await parsed_delete_request_non_return(
            client=self, 
            url=self._build_endpoint(DatastoreEndpoints.DELETE_RELEASE_SYS_REVIEWERS_DELETE),
            error_message="Failed to delete reviewer!", 
            params={"reviewer_id": reviewer_id},
        )
    
    async def add_dataset_reviewer(self, reviewer_id: str) -> None: 
        """Add a reviewer to the datastore.

        Parameters
        ----------
        reviewer_id : str
            Valid Id of a reviewer.
        """

        await parsed_post_request_none_return(
            client=self, 
            url=self._build_endpoint(DatastoreEndpoints.POST_RELEASE_SYS_REVIEWERS_ADD),
            error_message="Failed to add reviewer!", 
            params={},
            json_body={"reviewer_id": reviewer_id},
        )
    
    """

    #TODO: This has to be completed, once a new pydantic defined response model is setup
    for this endpoint.

    async def list_reviewers(self) -> EmptyResponse: 

        return await parsed_post_request(
            client=self, 
            url=self._build_endpoint(DatastoreEndpoints.POST_RELEASE_SYS_REVIEWERS_ADD),
            error_message="Failed to add reviewer!", 
            params={},
            json_body={"reviewer_id": reviewer_id},
            model=EmptyResponse
        )

    """

    async def approval_request(self, approval_request_payload: ReleaseApprovalRequest) -> ReleaseApprovalRequestResponse:
        """Submit a request for approval of dataset through the datastore.

        Parameters
        ----------
        approval_request_payload : ReleaseApprovalRequest
            An object that requires the dataset id, approver id and notes

        Returns
        -------
        ReleaseApprovalRequestResponse
            Contains details of the approval request.
        """

        return await parsed_post_request(
            client=self, 
            url=self._build_endpoint(DatastoreEndpoints.POST_RELEASE_APPROVAL_REQUEST),
            error_message="Failed to create an approval request!", 
            params={},
            json_body=py_to_dict(approval_request_payload),
            model=ReleaseApprovalRequestResponse
        )
        
    async def action_approval_request(self, action_approval_request_payload: ActionApprovalRequest) -> ActionApprovalRequestResponse:
        """Action an approval request from a dataset approval request via the datastore.

        Parameters
        ----------
        action_approval_request : ActionApprovalRequest
            The dataset id, your decision of approval and any extra information 
            you want to add (notes).

        Returns
        -------
        ActionApprovalRequestResponse
            The details of the approval action and the relevant dataset details.
        """
        
        return await parsed_put_request(
            client=self, 
            url=self._build_endpoint(DatastoreEndpoints.PUT_RELEASE_ACTION_APPROVAL_REQUEST),
            error_message="Failed to action approval request!", 
            params={},
            json_body=py_to_dict(action_approval_request_payload),
            model=ActionApprovalRequestResponse
        )
    
    # Admin endpoint. 
    # TODO - This has to be done.
    #  async def generate_config_file(self) -> None: 
    #    pass


class DatastoreClient(ClientService):

    review: DatasetReviewSubClient

    def __init__(self, auth: AuthManager, config: Config) -> None:
        """Initialise the DatastoreClient with authentication and configuration.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance.
        """
        self._auth = auth
        self._config = config

        self.review = DatasetReviewSubClient(auth = auth, config = config)

    def _build_endpoint(self, endpoint: DatastoreEndpoints) -> str:
        return self._config.datastore_api_endpoint + endpoint.value
        
    async def get_health_check(self) -> HealthCheckResponse:
        """
        Health check the API

        Returns:
            HealthCheckResponse: Response
        """
        return await parsed_get_request(
            client=self,
            url=self._build_endpoint(DatastoreEndpoints.GET_HEALTH_CHECK),
            error_message="Health check failed!",
            params={},
            model=HealthCheckResponse
        )
        
    async def validate_metadata(self, metadata_payload: CollectionFormat) -> StatusResponse:
        """Validates provided dataset info with the datastore API.

        Parameters
        ----------
        metadata_payload : CollectionFormat
            A structured format containing all necessary information to register a new dataset, including associations, 
            approvals, and dataset-specific information.


        Returns
        -------
        StatusResponse
            Response indicating whether your dataset metadata setup is valid and correct.
        """

        return await parsed_post_request_with_status(
            client = self, 
            url = self._build_endpoint(DatastoreEndpoints.POST_METADATA_VALIDATE_METADATA), 
            error_message="Dataset metadata validation failed!", 
            params = {},
            json_body=py_to_dict(metadata_payload),
            model = StatusResponse
        ) 
    
    async def update_metadata(self, handle_id: str, reason: str, metadata_payload: CollectionFormat) -> UpdateMetadataResponse:

        """Updates existing dataset metadata through datastore API.

        Returns
        -------
        UpdateMetadataResponse
            The updated metadata response from datastore.
        """

        return await parsed_post_request_with_status(
            client = self, 
            url = self._build_endpoint(DatastoreEndpoints.POST_REGISTER_UPDATE_METADATA),
            error_message="Dataset metadata update failed!", 
            params = {"handle_id": handle_id, "reason": reason}, 
            json_body=py_to_dict(metadata_payload), 
            model = UpdateMetadataResponse
        )
    
    async def revert_metadata(self, metadata_payload:RevertMetadata) -> StatusResponse:
        """Reverts the metadata for a dataset to a previous identified historical version.

        Parameters
        ----------
        metadata_payload : RevertMetadata
            The revert request, passed through to the registry API and requires
            dataset id, history id and reason for reverting.

        Returns
        -------
        StatusResponse
            Response indicating whether your dataset metadata setup is valid and correct.
        """

        return await parsed_put_request_with_status(
            client = self, 
            url = self._build_endpoint(DatastoreEndpoints.PUT_REGISTER_REVERT_METADATA),
            error_message= "Dataset revert metadata failed!",
            params={},
            json_body= py_to_dict(metadata_payload),
            model=StatusResponse
        )
    
    async def version_dataset(self, version_dataset_payload: VersionRequest) -> VersionResponse:
        """Creates a new versioning of an existing dataset within Provena 
           through the Datastore.

        Parameters
        ----------
        version_dataset_payload : VersionRequest
            The request which includes the item ID and reason for versioning.

        Returns
        -------
        VersionResponse
            Response of the versioning of the dataset, containing new version ID and 
            job session ID.
        """

        return await parsed_post_request(
            client = self, 
            url = self._build_endpoint(DatastoreEndpoints.POST_REGISTER_VERSION),
            error_message= "Dataset versioning failed!",
            params={},
            json_body= py_to_dict(version_dataset_payload),
            model=VersionResponse
        )
    
    async def list_datasets(self, list_request: NoFilterSubtypeListRequest) -> DatasetListResponse:
        """Gets datasets within the datastore in a paginated fashion.

        Parameters
        ----------
        list_request : NoFilterSubtypeListRequest
            Contains parameters for the specified sorting criteria, 
            optional pagination key and amount of records
            to fetch.

        Returns
        -------
        ListRegistryResponse
            Response of fetching datasets from datastore API.
        """

        return await parsed_post_request_with_status(
            client = self, 
            url = self._build_endpoint(DatastoreEndpoints.POST_REGISTRY_ITEMS_LIST),
            error_message="List fetching failed",
            params = {},
            json_body=py_to_dict(list_request),
            model = DatasetListResponse
        )
    

    async def generate_presigned_url(self, presigned_url: PresignedURLRequest) -> PresignedURLResponse: 
        """Generates a presigned url for an existing dataset within the datastore.

        Parameters
        ----------
        presigned_url : PresignedURLRequest
            Contains the dataset id + file path + length of expiry of URL.

        Returns
        -------
        PresignedURLResponse
            A response containing the presigned url.
        """

        return await parsed_post_request(
            client = self, 
            url = self._build_endpoint(DatastoreEndpoints.POST_REGISTRY_ITEMS_GENERATE_PRESIGNED_URL), 
            error_message= "Dataset presigned url generation failed!",
            params={},
            json_body=py_to_dict(presigned_url),
            model=PresignedURLResponse
        )

    async def generate_read_access_credentials(self, read_access_credentials: CredentialsRequest) -> CredentialResponse:
        """Creates a read-access for a certain subdirectory of the S3 bucket.

        Parameters
        ----------
        read_access_credentials : CredentialsRequest
            Contains the dataset id + console session URL required flag (boolean)

        Returns
        -------
        CredentialResponse
            The AWS credentials creating read level access into the subset of the bucket requested in the S3 location object.
        """

        return await parsed_post_request(
            client = self, 
            url = self._build_endpoint(DatastoreEndpoints.POST_REGISTRY_CREDENTIALS_GENERATE_READ_ACCESS_CREDENTIALS), 
            error_message= "Read access credential creation failed!",
            params={},
            json_body=py_to_dict(read_access_credentials),
            model=CredentialResponse
        )
    
    async def generate_write_access_credentials(self, write_access_credentials: CredentialsRequest) -> CredentialResponse:
        """Creates a write-access for a certain subdirectory of the S3 bucket. 

        Parameters
        ----------
        write_access_credentials : CredentialsRequest
            Contains the dataset id + console session URL required flag (boolean)

        Returns
        -------
        CredentialResponse
            The AWS credentials creating write level access into the subset of the bucket requested in the S3 location object.
        """

        return await parsed_post_request(
            client = self, 
            url = self._build_endpoint(DatastoreEndpoints.POST_REGISTRY_CREDENTIALS_GENERATE_WRITE_ACCESS_CREDENTIALS), 
            error_message= "Write access credential creation failed!",
            params={},
            json_body=py_to_dict(write_access_credentials),
            model=CredentialResponse
        )
    

    async def fetch_dataset(self, id: str) -> RegistryFetchResponse:
        """Fetches a dataset from the datastore based on the provided
        ID.

        Parameters
        ----------
        id : str
            The unique identifier of the dataset to be retrieved.
            For example: "10378.1/1451860"

        Returns
        -------
        RegistryFetchResponse
            A interactive python datatype of type RegistryFetchResponse
            containing the dataset details.

        Raises
        ------
        CustomTimeoutException
            Raised if the request times out.
        Exception
            General exception for handling unexpected errors.
        ValueError
            Raised if there is an issue in parsing the response into the expected model.
        """
        return await parsed_get_request_with_status(
            client=self,
            url=self._build_endpoint(
                DatastoreEndpoints.GET_REGISTRY_ITEMS_FETCH_DATASET),
            error_message=f"Failed to fetch dataset with id {id}...",
            params={"handle_id": id},
            model=RegistryFetchResponse
        )

    async def mint_dataset(self, dataset_info: CollectionFormat) -> MintResponse:
        """Creates a new dataset in the datastore with the provided dataset information.

        Parameters
        ----------
        dataset_info : CollectionFormat
            A structured format containing all necessary information to register a new dataset, including associations, 
            approvals, and dataset-specific information.

        Returns
        -------
        MintResponse
            A interactive python datatype of type MintResponse
            containing the newly created dataset details.

        Raises
        ------
        BadRequestException
            Raised if the server returns a 400 status code, indicating a bad request.
        CustomTimeoutException
            Raised if the request times out.
        Exception
            General exception for handling unexpected errors.
        ValueError
            Raised if there is an issue in parsing the response into the expected model.

        """
        return await parsed_post_request_with_status(
            client=self,
            url=self._build_endpoint(DatastoreEndpoints.POST_REGISTER_MINT_DATASET),
            error_message="Failed to mint the desired dataset...",
            params={},
            json_body=py_to_dict(dataset_info),
            model=MintResponse
        )
    
    

    

    






from provenaclient.modules.datastore import Datastore
from provenaclient.modules.search import Search
from provenaclient.modules.auth import Auth
from provenaclient.modules.link import Link
from provenaclient.modules.prov import Prov
from provenaclient.modules.registry import Registry
from provenaclient.modules.id_service import IDService
from provenaclient.modules.job_service import JobService


'''
Created Date: Monday June 17th 2024 +1000
Author: Peter Baker
-----
Last Modified: Monday June 17th 2024 4:45:39 pm +1000
Modified By: Peter Baker
-----
Description: Prov API L2 Client.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------

18-06-2024 | Peter Baker | Note that this layer does not provide any file IO capabilities - see L3
'''

from typing import List, cast
from provenaclient.auth.manager import AuthManager
from provenaclient.utils.config import Config
from enum import Enum
from provenaclient.utils.helpers import *
from provenaclient.clients.client_helpers import *
from provenaclient.models.general import HealthCheckResponse
from ProvenaInterfaces.ProvenanceAPI import LineageResponse,ModelRunRecord, RegisterModelRunResponse, RegisterBatchModelRunRequest, RegisterBatchModelRunResponse, ConvertModelRunsResponse
from ProvenaInterfaces.RegistryAPI import ItemModelRun

class ProvAPIEndpoints(str, Enum):
    """An ENUM containing the prov api endpoints."""
    
    # Completed
    POST_MODEL_RUN_REGISTER = "/model_run/register"
    POST_MODEL_RUN_REGISTER_BATCH = "/model_run/register_batch"
    GET_EXPLORE_UPSTREAM = "/explore/upstream"
    GET_EXPLORE_DOWNSTREAM = "/explore/downstream"
    GET_EXPLORE_SPECIAL_CONTRIBUTING_DATASETS = "/explore/special/contributing_datasets"
    GET_EXPLORE_SPECIAL_EFFECTED_DATASETS = "/explore/special/effected_datasets"
    GET_EXPLORE_SPECIAL_CONTRIBUTING_AGENTS = "/explore/special/contributing_agents"
    GET_EXPLORE_SPECIAL_EFFECTED_AGENTS = "/explore/special/effected_agents"
    GET_HEALTH_CHECK = "/"
    GET_BULK_GENERATE_TEMPLATE_CSV = "/bulk/generate_template/csv"
    POST_BULK_CONVERT_MODEL_RUNS_CSV = "/bulk/convert_model_runs/csv"
    GET_BULK_REGENERATE_FROM_BATCH_CSV = "/bulk/regenerate_from_batch/csv"

    
    
    # Not completed.
    GET_CHECK_ACCESS_CHECK_GENERAL_ACCESS = "/check-access/check-general-access"
    GET_CHECK_ACCESS_CHECK_ADMIN_ACCESS = "/check-access/check-admin-access"
    GET_CHECK_ACCESS_CHECK_READ_ACCESS = "/check-access/check-read-access"
    GET_CHECK_ACCESS_CHECK_WRITE_ACCESS = "/check-access/check-write-access"
 

class ProvAPIAdminEndpoints(str, Enum):
    """An ENUM containing the prov api admin endpoints."""
    
    # Completed
    GET_ADMIN_CONFIG = "/admin/config"
    POST_ADMIN_STORE_RECORD = "/admin/store_record"
    POST_ADMIN_STORE_RECORDS = "/admin/store_records"
    POST_ADMIN_STORE_ALL_REGISTRY_RECORDS = "/admin/store_all_registry_records"


    # Not completed yet, TODO.
    GET_ADMIN_SENTRY_DEBUG = "/admin/sentry-debug"
  
    
#L2 interface.


class ProvAdminClient(ClientService):

    def __init__(self, auth: AuthManager, config: Config) -> None:
        self._auth = auth
        self._config = config
    
    def _build_endpoint(self, endpoint: ProvAPIAdminEndpoints) -> str:
       return self._config.prov_api_endpoint + endpoint.value
    

    async def generate_config_file(self, required_only: bool) -> str:
        """Generates a nicely formatted .env file of the current required/non supplied properties 
        Used to quickly bootstrap a local environment or to understand currently deployed API.

        Parameters
        ----------
        required_only : bool, optional
            By default True
        """

        response = await validated_get_request(
            client=self, 
            url = self._build_endpoint(ProvAPIAdminEndpoints.GET_ADMIN_CONFIG),
            error_message=f"Failed to generate config file",
            params = {"required_only": required_only},
        )

        return response.text
    
    async def store_record(self, registry_record: ItemModelRun, validate_record: bool) -> StatusResponse:
        """An admin only endpoint which enables the reupload/storage of an existing completed provenance record.

        Parameters
        ----------
        registry_record : ItemModelRun
            The completed registry record for the model run.
        validate_record: bool
            Optional Should the ids in the payload be validated?, by default True

        Returns
        -------
        StatusResponse
            A status response indicating the success of the request and any other details.
        """

        return await parsed_post_request_with_status(
            client=self, 
            url = self._build_endpoint(ProvAPIAdminEndpoints.POST_ADMIN_STORE_RECORD),
            error_message=f"Failed to store record with display name {registry_record.display_name} and id {registry_record.id}",
            params = {"validate_record": validate_record},
            json_body=py_to_dict(registry_record),
            model = StatusResponse
        )
    
    async def store_multiple_records(self, registry_record: List[ItemModelRun], validate_record: bool) -> StatusResponse:
        """An admin only endpoint which enables the reupload/storage of an existing but multiple completed provenance record.

        Parameters
        ----------
        registry_record : List[ItemModelRun]
            List of the completed registry record for the model run validate_record
        validate_record: bool
            Optional Should the ids in the payload be validated?, by default True

        Returns
        -------
        StatusResponse
            A status response indicating the success of the request and any other details.
        """

        return await parsed_post_request_with_status(
            client=self, 
            url = self._build_endpoint(ProvAPIAdminEndpoints.POST_ADMIN_STORE_RECORDS),
            error_message=f"Failed to complete multiple store record request.",
            params = {"validate_record": validate_record},
            model = StatusResponse,
            json_body = cast(List[Dict[str, Any]], [py_to_dict(item) for item in registry_record])
        )
    
    async def store_all_registry_records(self, validate_record: bool) -> StatusResponse:
        """Applies the store record endpoint action across a list of ItemModelRuns '
           which is found by querying the registry model run list endpoint directly.

        Parameters
        ----------
        validate_record : bool
            Optional Should the ids in the payload be validated?, by default True


        Returns
        -------
        StatusResponse
            A status response indicating the success of the request and any other details.
        """

        return await parsed_post_request_with_status(
            client=self, 
            url = self._build_endpoint(ProvAPIAdminEndpoints.POST_ADMIN_STORE_ALL_REGISTRY_RECORDS),
            error_message=f"Failed to validate records.",
            params = {"validate_record": validate_record},
            json_body= None,
            model = StatusResponse
        )

    
class ProvClient(ClientService):

    admin: ProvAdminClient

    def __init__(self, auth: AuthManager, config: Config) -> None:
        """Initialises the REPLACEClient with authentication and configuration.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance.
        """
        self._auth = auth
        self._config = config

        self.admin = ProvAdminClient(auth=auth, config=config)

    def _build_endpoint(self, endpoint: ProvAPIEndpoints) -> str:
       return self._config.prov_api_endpoint + endpoint.value
    
    async def get_health_check(self) -> HealthCheckResponse:
        """Checks the health status of the PROV-API.

        Returns
        -------
        HealthCheckResponse
            Response containing the PROV-API health information.
        """

        return await parsed_get_request(
            client=self,
            url=self._build_endpoint(ProvAPIEndpoints.GET_HEALTH_CHECK),
            error_message="Health check failed!",
            params={},
            model=HealthCheckResponse
        )
    
    # Explore Lineage endpoints
    async def explore_upstream(self, starting_id: str, depth: int) -> LineageResponse:
        """Explores in the upstream direction (inputs/associations) 
        starting at the specified node handle ID. 
        The search depth is bounded by the depth parameter which has a default maximum of 100.

        Parameters
        ----------
        starting_id : str
            The ID of the entity to start at.
        depth : int, optional
            The depth to traverse in the upstream direction, by default 100.

        Returns
        -------
        LineageResponse
            A response containing the status, node count, and networkx serialised graph response.
        """

        return await parsed_get_request_with_status(
            client=self, 
            url = self._build_endpoint(ProvAPIEndpoints.GET_EXPLORE_UPSTREAM),
            error_message=f"Upstream query with starting id {starting_id} and depth {depth} failed!",
            params = {"starting_id": starting_id, "depth": depth},
            model = LineageResponse
        )

 
    async def explore_downstream(self, starting_id: str, depth: int) -> LineageResponse:
        """Explores in the downstream direction (inputs/associations) 
        starting at the specified node handle ID. 
        The search depth is bounded by the depth parameter which has a default maximum of 100.

        Parameters
        ----------
        starting_id : str
            The ID of the entity to start at.
        depth : int, optional
            The depth to traverse in the downstream direction, by default 100

        Returns
        -------
        LineageResponse
            A response containing the status, node count, and networkx serialised graph response.
        """
        
        return await parsed_get_request_with_status(
            client=self, 
            url = self._build_endpoint(ProvAPIEndpoints.GET_EXPLORE_DOWNSTREAM),
            error_message=f"Downstream query with starting id {starting_id} and depth {depth} failed!",
            params = {"starting_id": starting_id, "depth": depth},
            model = LineageResponse
        )
    
    async def get_contributing_datasets(self, starting_id: str, depth: int) -> LineageResponse:
        """Fetches datasets (inputs) which involved in a model run
        naturally in the upstream direction.

        Parameters
        ----------
        starting_id : str
            The ID of the entity to start at.
        depth : int, optional
            The depth to traverse in the upstream direction, by default 100

        Returns
        -------
        LineageResponse
            A response containing the status, node count, and networkx serialised graph response.
        """

        return await parsed_get_request_with_status(
            client=self, 
            url = self._build_endpoint(ProvAPIEndpoints.GET_EXPLORE_SPECIAL_CONTRIBUTING_DATASETS),
            error_message=f"Contributing datasets query with starting id {starting_id} and depth {depth} failed!",
            params = {"starting_id": starting_id, "depth": depth},
            model = LineageResponse
        )
    
    async def get_effected_datasets(self, starting_id: str, depth: int) -> LineageResponse:
        """Fetches datasets (outputs) which are derived from the model run
        naturally in the downstream direction.

        Parameters
        ----------
        starting_id : str
            The ID of the entity to start at.
        depth : int, optional
            The depth to traverse in the downstream direction, by default 100.

        Returns
        -------
        LineageResponse
            A response containing the status, node count, and networkx serialised graph response.
        """

        return await parsed_get_request_with_status(
            client=self, 
            url = self._build_endpoint(ProvAPIEndpoints.GET_EXPLORE_SPECIAL_EFFECTED_DATASETS),
            error_message=f"Effected datasets query with starting id {starting_id} and depth {depth} failed!",
            params = {"starting_id": starting_id, "depth": depth},
            model = LineageResponse
        )
    
    async def get_contributing_agents(self, starting_id: str, depth: int) -> LineageResponse:
        """Fetches agents (organisations or peoples) that are involved or impacted by the model run.
        naturally in the upstream direction.

        Parameters
        ----------
        starting_id : str
            The ID of the entity to start at.
        depth : int, optional
            The depth to traverse in the upstream direction, by default 100.

        Returns
        -------
        LineageResponse
            A response containing the status, node count, and networkx serialised graph response.
        """

        return await parsed_get_request_with_status(
            client=self, 
            url = self._build_endpoint(ProvAPIEndpoints.GET_EXPLORE_SPECIAL_CONTRIBUTING_AGENTS),
            error_message=f"Contributing agents query with starting id {starting_id} and depth {depth} failed!",
            params = {"starting_id": starting_id, "depth": depth},
            model = LineageResponse
        )
    
    async def get_effected_agents(self, starting_id: str, depth: int) -> LineageResponse:
        """Fetches agents (organisations or peoples) that are involved or impacted by the model run.
        naturally in the downstream direction.

        Parameters
        ----------
        starting_id : str
            The ID of the entity to start at.
        depth : int, optional
            The depth to traverse in the downstream direction, by default 100.

        Returns
        -------
        LineageResponse
            A response containing the status, node count, and networkx serialised graph response.
        """

        return await parsed_get_request_with_status(
            client=self, 
            url = self._build_endpoint(ProvAPIEndpoints.GET_EXPLORE_SPECIAL_EFFECTED_AGENTS),
            error_message=f"Effected agents query with starting id {starting_id} and depth {depth} failed!",
            params = {"starting_id": starting_id, "depth": depth},
            model = LineageResponse
        )
    

    # Model run endpoints.

    async def register_batch_model_runs(self, model_run_batch_payload:RegisterBatchModelRunRequest) -> RegisterBatchModelRunResponse:
        """This function allows you to register multiple model runs in one go (batch) asynchronously.

        Note: You can utilise the returned session ID to poll on 
        the JOB API to check status of the model run registration(s).

        Parameters
        ----------
        batch_model_run_payload : RegisterBatchModelRunRequest
            A list of model runs (ModelRunRecord objects)

        Returns
        -------
        RegisterBatchModelRunResponse
            The job session id derived from job-api for the model-run batch. 
        """

        return await parsed_post_request(
            client = self, 
            url = self._build_endpoint(ProvAPIEndpoints.POST_MODEL_RUN_REGISTER_BATCH),
            error_message=f"Model run batch registration failed!",
            params = {},
            json_body=py_to_dict(model_run_batch_payload),
            model = RegisterBatchModelRunResponse
        )    

    async def register_model_run(self, model_run_payload: ModelRunRecord) -> RegisterModelRunResponse:
        """Asynchronously registers a single model run.

        Note: You can utilise the returned session ID to poll on 
        the JOB API to check status of the model run registration.

        Parameters
        ----------
        model_run_payload : ModelRunRecord
            Contains information needed for the 
            model run such as workflow template,
            inputs, outputs, description etc. 

        Returns
        -------
        RegisterModelRunResponse
            The job session id derived from job-api for the model-run. 
        """

        return await parsed_post_request(
            client = self, 
            url = self._build_endpoint(ProvAPIEndpoints.POST_MODEL_RUN_REGISTER), 
            error_message=f"Model run registration failed!", 
            params = {},
            json_body=py_to_dict(model_run_payload),
            model = RegisterModelRunResponse
        )
    

    # CSV template tools endpoints
    async def generate_csv_template(self, workflow_template_id: str) -> str:
        """Generates a model run csv template to be utilised 
        for creating model runs through csv format.

        Parameters
        ----------
        workflow_template_id : str
            An ID of a created and existing model run workflow template.
        """

        response = await validated_get_request(
            client=self, 
            url = self._build_endpoint(ProvAPIEndpoints.GET_BULK_GENERATE_TEMPLATE_CSV),
            error_message=f"Failed to generate CSV file",
            params = {"workflow_template_id": workflow_template_id},
        )


        return response.text
    
    async def convert_model_runs_to_csv(self, csv_file_contents: str) -> ConvertModelRunsResponse:
        """Reads a CSV file, and it's defined model run contents
        and lodges a model run.

        Parameters
        ----------
        csv_file_contents : str
            Contains the model run contents.

        Returns
        -------
        ConvertModelRunsResponse
            Returns the model run information in an interactive python
            datatype.
        """

        # Convert string to bytes. 
        try:
            model_run_content_encoded: ByteString = csv_file_contents.encode("utf-8")
        except Exception as e:
            raise Exception(f"Exception has occurred while encoding model run content: {e}")

        # The csv file object to be used for httpx post requests
        # A dictionary representing file(s) to be uploaded with the
        # request. Each key in the dictionary is the name of the form field for the file according,
        # to API specifications. For Provena it's "csv_file" and and the value 
        # is a tuple of (filename, filedata, MIME type / media type).  
        csv_file: HttpxFileUpload = {"csv_file": ("upload.csv", model_run_content_encoded, "text/csv")}
    
        return await parsed_post_request_with_status(
            client=self, 
            url=self._build_endpoint(ProvAPIEndpoints.POST_BULK_CONVERT_MODEL_RUNS_CSV),
            error_message="Failed to generate CSV file",
            files=csv_file,
            json_body=None,
            params={},
            model=ConvertModelRunsResponse
        )
    
    async def regenerate_csv_from_model_run_batch(self, batch_id: str) -> str:
        """Regenerate/create a csv file containing model 
        run information from a model run batch job.

        The batch id must exist in the system.

        Parameters
        ----------
        batch_id : str
            Obtained from creating a batch model run.
        """

        response = await validated_get_request(
            client=self, 
            url = self._build_endpoint(ProvAPIEndpoints.GET_BULK_REGENERATE_FROM_BATCH_CSV),
            error_message=f"Failed to generate CSV file from batch_id {batch_id}",
            params = {"batch_id": batch_id},
        )

        return response.text







        



'''
Created Date: Tuesday May 28th 2024 +1000
Author: Peter Baker
-----
Last Modified: Tuesday May 28th 2024 11:18:32 am +1000
Modified By: Peter Baker
-----
Description: Defines the primary config object for the Provena Client - notably the endpoints.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

from pydantic import BaseModel
from typing import Optional

def optional_override_prefixor(domain: str, prefix: str, override: Optional[str]) -> str:
    """

    Helper function which performs the API prefixing while being override aware. 

    If override supplied - uses it directly. 

    If not, prefixes with protocol and prefix suitably.

    Args:
        domain (str): The domain base
        prefix (str): The API prefix
        override (Optional[str]): The optional override

    Returns:
        str: The URL
    """
    if override is not None:
        return override
    else:
        if prefix != "":
            return f"https://{prefix}.{domain}"
        else:
            return f"https://{domain}"


class APIOverrides(BaseModel):
    datastore_api_endpoint_override: Optional[str] = None
    auth_api_endpoint_override: Optional[str] = None
    registry_api_endpoint_override: Optional[str] = None
    prov_api_endpoint_override: Optional[str] = None
    search_api_endpoint_override: Optional[str] = None
    search_service_endpoint_override: Optional[str] = None
    handle_service_api_endpoint_override: Optional[str] = None
    jobs_service_api_endpoint_override: Optional[str] = None
    keycloak_endpoint_override: Optional[str] = None

class EndpointConfig(BaseModel):
    domain: str
    # What is the auth realm name?
    realm_name: str
    api_overrides: APIOverrides

class Config():

    def __init__(self, domain: str, realm_name: str, api_overrides: APIOverrides = APIOverrides()) -> None:
        """Creates a EndpointConfig object that holds relevant Provena instance information
        and possible overrides if provided.

        Parameters
        ----------
        domain : str
            The current domain that the Provena instance is deployed in.
        realm_name : str
            Your keycloak realm name.
        api_overrides : APIOverrides, optional
            Provide any overrides to certain API endpoints if you wish, by default APIOverrides() with all overrides set to None.
        """

        # the unpopulated environment
        self._api_config: EndpointConfig = EndpointConfig(domain=domain, realm_name=realm_name, api_overrides=api_overrides)
    
    # Property methods to retrieve different API endpoints. 

    @property
    def search_api_endpoint(self) -> str:
        """Generate the search api endpoint based on the 
        provided domain, prefix and possible override value. 

        Returns
        -------
        str
            A string containing the search api endpoint.
        """
        
        return optional_override_prefixor(
            domain=self._api_config.domain,
            prefix="search",
            override= self._api_config.api_overrides.search_api_endpoint_override if self._api_config.api_overrides else None
        )

    @property
    def jobs_service_api_endpoint(self) -> str:
        """Generate the job api endpoint based on the 
        provided domain, prefix and possible override value. 

        Returns
        -------
        str
            A string containing the job api endpoint.
        """

        return optional_override_prefixor(
            domain=self._api_config.domain,
            prefix="job-api",
            override=self._api_config.api_overrides.jobs_service_api_endpoint_override if self._api_config.api_overrides else None
        )

    @property
    def handle_service_api_endpoint(self) -> str:
        """Generate the handle_service api endpoint based on the 
        provided domain, prefix and possible override value. 

        Returns
        -------
        str
            A string containing the handle_service api endpoint.
        """

        return optional_override_prefixor(
            domain=self._api_config.domain,
            prefix="handle",
            override=self._api_config.api_overrides.handle_service_api_endpoint_override if self._api_config.api_overrides else None
        )

    @property
    def search_service_endpoint(self) -> str:
        """Generate the search_service api endpoint based on the 
        provided domain, prefix and possible override value. 

        Returns
        -------
        str
            A string containing the search_service api endpoint.
        """

        return optional_override_prefixor(
            domain=self._api_config.domain,
            prefix="search-service",
            override=self._api_config.api_overrides.search_service_endpoint_override if self._api_config.api_overrides else None
        )

    @property
    def auth_api_endpoint(self) -> str:
        """Generate the auth api endpoint based on the 
        provided domain, prefix and possible override value. 

        Returns
        -------
        str
            A string containing the auth api endpoint.
        """

        return optional_override_prefixor(
            domain=self._api_config.domain,
            prefix="auth-api",
            override=self._api_config.api_overrides.auth_api_endpoint_override if self._api_config.api_overrides else None
        )

    @property
    def prov_api_endpoint(self) -> str:
        """Generate the prov api endpoint based on the 
        provided domain, prefix and possible override value. 

        Returns
        -------
        str
            A string containing the prov api endpoint.
        """

        return optional_override_prefixor(
            domain=self._api_config.domain,
            prefix="prov-api",
            override=self._api_config.api_overrides.prov_api_endpoint_override if self._api_config.api_overrides else None
        )

    @property
    def datastore_api_endpoint(self) -> str:
        """Generate the datastore api endpoint based on the 
        provided domain, prefix and possible override value. 

        Returns
        -------
        str
            A string containing the datastore api endpoint.
        """

        return optional_override_prefixor(
            domain=self._api_config.domain,
            prefix="data-api",
            override=self._api_config.api_overrides.datastore_api_endpoint_override if self._api_config.api_overrides else None
        )

    @property
    def registry_api_endpoint(self) -> str:
        """Generate the registry api endpoint based on the 
        provided domain, prefix and possible override value. 

        Returns
        -------
        str
            A string containing the registry api endpoint.
        """

        return optional_override_prefixor(
            domain=self._api_config.domain,
            prefix="registry-api",
            override=self._api_config.api_overrides.registry_api_endpoint_override if self._api_config.api_overrides else None
        )

    @property
    def keycloak_endpoint(self) -> str:
        """Generate the keycloak realm endpoint using domain, realm_name and possible 
        override value. 

        Returns
        -------
        str
            A string containing the keycloak realm endpoint.
        """

        endpoint = ""
        if self._api_config.api_overrides.keycloak_endpoint_override is not None:
            endpoint = self._api_config.api_overrides.keycloak_endpoint_override
        else:
            endpoint = f"https://auth.{self._api_config.domain}/auth/realms/{self._api_config.realm_name}"
        return endpoint
    

    
    



'''
Created Date: Thursday June 6th 2024 +1000
Author: Peter Baker
-----
Last Modified: Thursday June 6th 2024 1:39:55 pm +1000
Modified By: Peter Baker
-----
Description: Set of route maps for the Registry API used to help generate routes from subtype + Action combinations.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

from enum import Enum
from typing import Dict
from ProvenaInterfaces.RegistryModels import ItemSubType
 
class RegistryAction(str, Enum):
    FETCH = "FETCH"
    UPDATE = "UPDATE"
    CREATE = "CREATE"
    LIST = "LIST"
    SEED = "SEED"
    VERSION = "VERSION"
    SCHEMA = "SCHEMA"
    UI_SCHEMA = "UI_SCHEMA"
    VALIDATE = "VALIDATE"
    AUTH_EVALUATE = "AUTH_EVALUATE"
    AUTH_CONFIGURATION = "AUTH_CONFIGURATION"
    AUTH_ROLES = "AUTH_ROLES"
    LOCK = "LOCK"
    UNLOCK = "UNLOCK"
    REVERT = "REVERT"
    LOCK_HISTORY = "LOCK_HISTORY"
    DELETE = "DELETE"
 
# Define the mappings
action_postfixes: Dict[RegistryAction, str] = {
    RegistryAction.FETCH: "/fetch",
    RegistryAction.UPDATE: "/update",
    RegistryAction.REVERT: "/revert",
    RegistryAction.CREATE: "/create",
    RegistryAction.LIST: "/list",
    RegistryAction.SEED: "/seed",
    RegistryAction.VERSION: "/version",
    RegistryAction.SCHEMA: "/schema",
    RegistryAction.UI_SCHEMA: "/ui_schema",
    RegistryAction.VALIDATE: "/validate",
    RegistryAction.AUTH_EVALUATE: "/auth/evaluate",
    RegistryAction.AUTH_CONFIGURATION: "/auth/configuration",
    RegistryAction.AUTH_ROLES: "/auth/roles",
    RegistryAction.LOCK: "/locks/lock",
    RegistryAction.UNLOCK: "/locks/unlock",
    RegistryAction.LOCK_HISTORY: "/locks/history",
    RegistryAction.DELETE: "/delete"
}
 
subtype_route_prefixes: Dict[ItemSubType, str] = {
    ItemSubType.MODEL_RUN: "/registry/activity/model_run",
    ItemSubType.ORGANISATION: "/registry/agent/organisation",
    ItemSubType.PERSON: "/registry/agent/person",
    ItemSubType.MODEL: "/registry/entity/model",
    ItemSubType.MODEL_RUN_WORKFLOW_TEMPLATE: "/registry/entity/model_run_workflow",
    ItemSubType.DATASET_TEMPLATE: "/registry/entity/dataset_template",
    ItemSubType.DATASET: "/registry/entity/dataset",
    ItemSubType.CREATE: "/registry/activity/create",
    ItemSubType.VERSION: "/registry/activity/version",
    ItemSubType.STUDY: "/registry/activity/study",
}
 
# Function to get the endpoint URL
def subtype_action_to_endpoint(base: str, action: RegistryAction, item_subtype: ItemSubType) -> str:
    """
    
    Generates a fully formed registry URL given the below params

    Args:
        base (str): The URL base e.g. https://registry.provena.com
        action (RegistryAction): The action to be taken, see enum
        item_subtype (ItemSubType): The subtype to get the route for

    Returns:
        str: The fully formed URL
    """
    subtype_prefix = subtype_route_prefixes[item_subtype]
    action_postfix = action_postfixes[action]
    return f"{base}{subtype_prefix}{action_postfix}"


# from provenaclient.auth.manager import AuthManager
# from provenaclient.utils.config import Config
# from provenaclient.utils.http_client import HttpClient
# from enum import Enum
# from provenaclient.utils.helpers import *
# from provenaclient.clients.client_helpers import *
#
#
# class REPLACEEndpoints(str, Enum):
#    """An ENUM containing the REPLACE endpoints.
#    """
#    REPLACE_ENDPOINT: str = "TODO"
#
# L2 interface.
#
#
# class REPLACEClient(ClientService):
#    def __init__(self, auth: AuthManager, config: Config) -> None:
#        """Initialises the REPLACEClient with authentication and configuration.
#
#        Parameters
#        ----------
#        auth : AuthManager
#            An abstract interface containing the user's requested auth flow method.
#        config : Config
#            A config object which contains information related to the Provena instance.
#        """
#        self._auth = auth
#        self._config = config

#    def _build_endpoint(self, endpoint: REPLACEEndpoints) -> str:
#        return self._config.replace_api_endpoint + endpoint.value
#
#    async def example_method(self) -> BaseModel:
#        return await parsed_get_request(
#            client=self,
#            url=self._build_endpoint(REPLACEEndpoints.TODO),
#            error_message="TODO",
#            params={},
#            model=BaseModel
#        )
